import numpy as _np

def kurt(y:_np.ndarray | list):
	"""
	Computes excess kurtosis. \n
	y: ndarray / list.
	"""
	n = len(y)

	assert n >= 4, "list/ndarray must have at least 4 elements"
	assert isinstance(y, list) or isinstance(y, _np.ndarray), "list/ndarray expected"
	
	Arr = y
	if(isinstance(y, list)):
		Arr = _np.asfarray(y)

	avg = _np.mean(Arr)
	stdev = _np.std(Arr, ddof=1)

	s = 0
	for Num in y:
		s += (Num - avg)**4

	Kurt = (s/n)/(stdev)**4 - 3

	return float(Kurt)


from cmath import inf
import wx
import string



class NumTextCtrl(wx.TextCtrl):
	"""
	A text ctrl that only allows numeric entries
	Decimal separator used is .
	"""
	def __init__(self, parent, id = wx.ID_ANY, val:str = wx.EmptyString, minval:float = -inf, maxval:float = inf):
		wx.TextCtrl.__init__(self, parent, id)

		self.m_Min = minval
		self.m_Max = maxval
		self.m_InitVal = val

		if(minval>-inf and maxval<inf):
			self.SetToolTip(self.ToRange(minval, maxval))

		if(val != wx.EmptyString):
			try:
				numVal=float(val)
				self.SetValue(val)
			except ValueError as ve:
				pass

		self.Bind(wx.EVT_CHAR, self.OnChar)
		self.Bind(wx.EVT_TEXT, self.OnText)
		self.Bind( wx.EVT_KILL_FOCUS, self.OnKillFocus )
	

	def ToRange(self, minval:float, maxval:float):
		return "Expected range [" + str(minval) + "," + str(maxval)+ "]"

	
	def OnKillFocus(self, event):
		val = self.GetValue()
		if(val != wx.EmptyString):
			try:
				numVal=float(val)
			except ValueError as ve:
				self.SetValue("")
		
		event.Skip()


	def OnText(self, event):
		if self.GetValue() == "":
			event.Skip()
			return
			
		NumVal = float(self.GetValue())
		if(NumVal<self.m_Min or NumVal>self.m_Max):
			wx.MessageBox(self.ToRange(self.m_Min, self.m_Max))

			#reset the value so that user will not be bugged when trying to recover from a mistake
			self.SetValue(self.m_InitVal)
		else:
			event.Skip()


	def OnChar(self, event):
		key = event.GetKeyCode()
		val = self.GetValue()
		if key == wx.WXK_NONE:
			pass 

		elif chr(key) in string.digits:
			event.Skip()

		elif key==wx.WXK_DELETE or key==wx.WXK_BACK or key==wx.WXK_HOME:
			event.Skip()
		
		elif key==wx.WXK_LEFT or key == wx.WXK_RIGHT:
			event.Skip()
		
		#dont allow duplicate separators
		elif chr(key)=='.' and '.' not in val:
			event.Skip()
		
		elif (chr(key) =='E' or chr(key)=='e') and ('E' not in val) and ('e' not in val):
			#if there is no character then E or e not make any sense
			if(val != wx.EmptyString):
				#if first character is minus then we need at least 2 characters and second one must be digit
				if(val[0]=='-'):
					if(len(val)>=2 and val[1] in string.digits):
						event.Skip()
				else:
					#if first character is not minus and E or e not already entered, allow it
					event.Skip()

		
		#only allow minus at the beginning
		elif chr(key)=='-':
			if(val == wx.EmptyString):
				event.Skip()
			else:
				HasE= ('E' in val) or ('e' in val)
				#if there is already minus at the beginning dont allow to add more
				if val[0] == '-' and not HasE:
					self.SetValue(val[0:])
				else:
					if HasE:
						event.Skip()
					else:
						self.SetValue('-' + val)
		
		elif chr(key) == ',':
			wx.MessageBox("Use decimal point (.) as the decimal separator")


import numpy as np
from dataclasses import dataclass

from ..stats import pf



__all__= ['aov2_results', 'aov']




def _parsedata(y, x1, x2, v1, v2):
	"""
	parse the data that comes in the format of y, x1, x2 to
	columns of x1 (v1 columns) and rows of x2 (v1 columns)

	The formatted data looks like how Excel accepts data for 2-factor ANOVA
	"""
	Tbl = []
	for i in range(len(v2)):
		Tbl.append(list())
		for j in range(len(v1)):
			Tbl[i].append(list())

	for k in range(len(y)):
		i, j = -1, -1

		while (True):
			i += 1 
			if x2[k] == v2[i]: break

		while (True):
			j +=1
			if x1[k] == v1[j]: break
		
		Tbl[i][j].append(y[k])
	
	return Tbl


def _averageMatrix(Tbl, v2):
	"""
	contains average value of each cell's data list
	"""
	MatAverage = []
	for i in range(len(v2)):
		arr = np.array(Tbl[i]).transpose()
		v = arr.mean(axis = 0)
		MatAverage.append(v.tolist())
	return MatAverage






@dataclass
class aov2_results():
	DFError:int; DFFact1:int; DFFact2:int; DFinteract:int
	FvalFact1:float; FvalFact2:float; Fvalinteract:float
	MSError:float; MSFact1:float; MSFact2:float; MSinteract:float
	pvalFact1:float; pvalFact2:float; pvalinteract:float
	SSError:float; SSFact1:float; SSFact2:float; SSinteract:float

	Residuals:list
	Fits:list



def aov2(y, x1, x2):
	"""
	Performs 2-way ANOVA

	## Input
	y: Responses \n
	x1, x2: factors
	"""
	X1 = x1
	if isinstance(x1, list):
		X1 = np.asfarray(x1)
	
	X2 = x2
	if isinstance(x2, list):
		X2 = np.asfarray(x2)
	
	YY = y
	if isinstance(y, list):
		YY = np.asfarray(y)

	assert len(X1)>= 3, "x1 must have at least 3 elements"
	assert len(X2) == len(X1), "x1 and x2 must have same size"
	assert len(X1) == len(YY), "x1 and y must have same size"

	assert np.issubdtype(X1.dtype, np.number), "x1 must contain only numbers"
	assert np.issubdtype(X2.dtype, np.number), "x2 must contain only numbers"
	assert np.issubdtype(YY.dtype, np.number), "y must contain only numbers"

	v1 = np.unique(X1)
	v2 = np.unique(X2)

	Tbl = _parsedata(YY, X1, X2, v1, v2)
	MatAverage = _averageMatrix(Tbl, v2)

	SSerror = 0
	Residuals, Fits = [], []

	for i in range(len(v2)):
		for j in range(len(v1)):
			sz = len(Tbl[i][j])
			for k in range(sz):
				Err = Tbl[i][j][k] - MatAverage[i][j]
				SSerror += Err**2.0

				Residuals.append(Err)
				Fits.append(MatAverage[i][j])

	MatAverage = np.array(MatAverage)
	GrandMean = MatAverage.mean()
	Nreplicate = len(Tbl[0][0])
	DFerror = (Nreplicate - 1) * (len(v1)) * (len(v2))
	MSerror = SSerror / DFerror

	meanj = MatAverage.mean(axis = 0)
	meani = MatAverage.transpose().mean(axis = 0)

	DFFact1, DFFact2 = len(v1) - 1, len(v2) - 1
	SSFact1 = np.sum((meanj - GrandMean)**2) * Nreplicate * len(v2)
	SSFact2 = np.sum((meani - GrandMean)**2) * Nreplicate * len(v1)
	MSFact1, MSFact2 = SSFact1 / DFFact1, SSFact2 / DFFact2

	SSinteract = 0
	for i in range(len(v2)):
		for j in range(len(v1)):
			SSinteract += (MatAverage[i, j] - meanj[j] - meani[i] + GrandMean)**2

	SSinteract *= Nreplicate
	DFinteract = DFFact1 * DFFact2
	MSinteract = SSinteract / DFinteract

	FvalFact1, FvalFact2, Fvalinteract = MSFact1 / MSerror, MSFact2 / MSerror, MSinteract / MSerror
	pvalFact1 = 1.0 - pf(FvalFact1, DFFact1, DFerror)
	pvalFact2 = 1.0 - pf(FvalFact2, DFFact2, DFerror)
	pvalinteract = 1.0 - pf(Fvalinteract, DFinteract, DFerror)

	return aov2_results(
		DFError=DFerror, DFFact1=DFFact1, DFFact2=DFFact2,DFinteract=DFinteract,
		FvalFact1=FvalFact1, FvalFact2 =FvalFact2, Fvalinteract = Fvalinteract,
		MSError = MSerror, MSFact1 = MSFact1, MSFact2 = MSFact2, MSinteract= MSinteract,
		pvalFact1 = pvalFact1, pvalFact2 = pvalFact2, pvalinteract= pvalinteract,
		SSError = SSerror, SSFact1 = SSFact1, SSFact2 = SSFact2, SSinteract=SSinteract,

		Fits=Fits,
		Residuals=Residuals
	)


"""
this file/module intended to serve as a temporary storage for variables <br>
those are exported from other parts <br>

How it works: <br>
1) A list variable is created and named as myvar <br>
2) To access the variable from other parts, issue the following command <br>
>> from scisuit.vars import myvar <br>
                                          <br>
now myvar can be used as a variable.   <br>
                                          <br>

P.S: Needless to say, myvar will not be visible as a text

"""


import math as _math


def colnum2labels(num:int):
	"""
	Finds the corresponding letter to num.
	For example A=1, B=2, Z=26, AA=27, AB=28, ACR=772, BCK=1441
	"""

	assert num>0, "num must be greater than 0"
	if(type(num) != int):
		raise TypeError("num type must be int")

	Str =""

	NChars = int(_math.ceil(_math.log(num) / _math.log(26.0)))
	if NChars == 1:
		modular = num % 26
		if (modular == 0):
			modular = 26
		Str = chr(65 + modular - 1)
		return Str
	
	val = num
	tbl = []
	for i in range(NChars-1):
		val /= 26
		tbl.append(chr(65 + int(val % 26) - 1))
	
	tbl.reverse()
	Str = "".join(tbl)
	Str += chr(65 + int(num % 26) - 1)

	return Str



from .path import parent_path
from .pkg import pkg_installed, assert_pkg



import sqlite3 as _sql

from ..util import parent_path as _parent_path
from .. import linearinterp as _linearinterp


__all__ = ['ThermoPhysical', 'SaturatedRefrigerant', 'SuperHeatedRefrigerant']


"""
This module depends on existence of the two databases: refrigerants.db and thermophysical.db

Properties are searched in the above-mentioned respective database, bracketed and then 
the intended value is computed using linear-interpolation(s).
"""



class Refrigerant():
	"""
	Base class for thermodynamic properties of refrigerants
	"""
	s_DataBasePath = _parent_path(__file__) / "refrigerants.db"
	
	def __init__(self) -> None: 
		self.m_Connection = _sql.connect(self.s_DataBasePath) 

		#-1:Compressed, 0:saturated, 1:superheated
		self.m_FluidState = None

		self.m_DBTable = None

	
	def __del__( self ):
		self.m_Connection.close()
	

	def Init(self, FluidName:str)->None:
		self.m_FluidName = FluidName
		cursor = self.m_Connection.cursor()

		"""
		Check if the parameter FluidName is valid
		Note that the columns in the database is configured as
		COLLATE NOCASE, therefore search is not case-sensitive
		"""
		QueryString = "SELECT * FROM MAINTABLE where NAME=?"
		rows = cursor.execute(QueryString , (FluidName,)).fetchall()

		#a name like "R" or "R1" will match more than 1
		if(len(rows)>1):
			raise ValueError("More than 1 fluid matched the name:" + FluidName)

		DBTableColPos = 2 #saturated, compressed
		if(self.m_FluidState == 1):
			DBTableColPos = 3
		
		if(len(rows) == 0):
			QueryString = "SELECT * FROM MAINTABLE where ALTERNATIVE=?"
			rows = cursor.execute(QueryString , (FluidName,)).fetchall()

			#all options exhausted, raise an error
			if(len(rows)==0):
				raise ValueError(FluidName + " did not match any")
            
			self.m_DBTable = rows[0][DBTableColPos]
        
		#len(rows) ==1
		else:
			self.m_DBTable = rows[0][DBTableColPos]



	def GetFluidNames(self):
		QueryString = "SELECT name, alternative FROM MAINTABLE"
		rowList = self.m_Connection.cursor().execute(QueryString , []).fetchall()
		
		return rowList
    



class SaturatedRefrigerant(Refrigerant):

	def __init__(self, FluidName:str) -> None:
		"""
		FluidName: Name of the fluid
		"""
		super().__init__() 

		self.m_FluidState = 0
		super().Init(FluidName)
		

	def search(self, PropertyName:str, QueryValue:float, Sort = True): 
		return _searchOrderedTable(self.m_Connection, self.m_DBTable, PropertyName, QueryValue, Sort) 



class SuperHeatedRefrigerant(Refrigerant):

	def __init__(self, FluidName:str) -> None:
		"""
		FluidName: Name of the fluid
		"""
		super().__init__() 

		self.m_FluidState = 1 
		super().Init(FluidName)
	

	def _BracketPressure(self, P:float):
		cursor = self.m_Connection.cursor()

		QueryString = "SELECT min(P), max(P) FROM " + self.m_DBTable
		MinMax = cursor.execute(QueryString , []).fetchall()
		Pmin, Pmax = MinMax[0][0], MinMax[0][1]

		if(not (Pmin<P and P<Pmax)):
			raise ValueError("Pressure range: ["+ str(Pmin) + ", " + str(Pmax) + "]" )
		
		#Pressure is within limits, but where?
		strQuery="SELECT DISTINCT P FROM " + self.m_DBTable + " WHERE P<="+str(P) +" ORDER BY P DESC LIMIT 1"
		row = cursor.execute(strQuery , []).fetchall()
		PL =row[0][0]
		
		strQuery="SELECT DISTINCT P FROM " + self.m_DBTable + " WHERE P>="+ str(P) +" LIMIT 1"
		row = cursor.execute(strQuery , []).fetchall()
		PH =row[0][0]
		
		return PL, PH

	
	def _BracketProperty(self, P:float, Name:str , Value:float):
		"""
		given a pressure value, finds lower and upper range of property
		"""

		cursor = self.m_Connection.cursor()
		
		QueryString = "SELECT min("+ Name +"), max(" + Name +") FROM " + self.m_DBTable + " WHERE P=?"
		MinMax = cursor.execute(QueryString , [P]).fetchall()
		PropertyMin, PropertyMax = MinMax[0][0], MinMax[0][1]

		if(not (PropertyMin < Value and Value < PropertyMax)):
			raise ValueError("At P="+ str(P) + " kPa, " + Name + " has range: ["+ str(PropertyMin) + ", " + str(PropertyMax) + "]" )
		
		"""
		Assumption is made that at a given pressure the property values
		are monotonically increasing (which is the case for T, V, H, S)
		"""
 
		strQuery="SELECT "+ Name +" FROM "+ self.m_DBTable + " WHERE P=? AND " + Name + "<=? ORDER BY " + Name + " DESC LIMIT 1"
		row = cursor.execute(strQuery , [P,Value]).fetchall()
		LowerRange =row[0][0]

		strQuery="SELECT " + Name +" FROM "+ self.m_DBTable + " WHERE P=? AND "+ Name +">=? LIMIT 1"
		row = cursor.execute(strQuery , [P, Value]).fetchall()
		UpperRange = row[0][0]

		return LowerRange, UpperRange


	def _FindProperties(self, P: float, Name:str , Value:float):
		"""
		Given a pressure value and a property, i.e. temperature
		the function returns the properties at the given pressure and temperature
		
		Note that the pressure here is the lower or upper value of the bracketed pressure
		and NOT the exact pressure value requested by user
		
		Therefore for any search this function needs to be called twice: at lower and upper ranges of pressure
		"""
		cursor = self.m_Connection.cursor()

		PropLow, PropHigh = self._BracketProperty(P, Name, Value)

		AllFieldNames:list = _GetFieldNames(self.m_Connection, self.m_DBTable)
		AllFieldNames.remove("P") #P is always known
		AllFieldNames.remove(Name.capitalize()) #property is known as well

		#As of this point AllFieldNames contains only the properties' names we are after
		JoinedFieldNames = ",".join(AllFieldNames)

		#Find properties at Pressure and lower range of the property 
		strQuery="SELECT " + JoinedFieldNames +" FROM "+ self.m_DBTable + " WHERE P=? AND "+ Name +"=?"
		
		"""
		Database table contains only fields of P, T, V, H, S
		Therefore once two properties are known, i.e., P and T,
		there remains 3 properties which are named arbitrarily as A, B, C
		"""
		
		row = cursor.execute(strQuery , [P, PropLow]).fetchall()
		Alow, Blow, Clow =row[0][0], row[0][1], row[0][2]

		#Find properties at lPressure and Upper range of the property 
		row = cursor.execute(strQuery , [P, PropHigh]).fetchall()
		Aup, Bup, Cup =row[0][0], row[0][1], row[0][2]

		A = _linearinterp(PropLow, Alow, PropHigh, Aup, Value)
		B = _linearinterp(PropLow, Blow, PropHigh, Bup, Value)
		C = _linearinterp(PropLow, Clow, PropHigh, Cup, Value)

		return A, B, C


	def search_PT(self, P:float, T:float):
		"""
		search for the values at a given pressure (kPa) and temperature (celcius)
		"""
		cursor = self.m_Connection.cursor()

		PL, PH = self._BracketPressure(P)

		Vlow, Hlow, Slow = self._FindProperties(PL, "T", T)
		Vup, Hup, Sup = self._FindProperties(PH, "T", T)

		V = _linearinterp(PL, Vlow, PH, Vup, P)
		H = _linearinterp(PL, Hlow, PH, Hup, P)
		S = _linearinterp(PL, Slow, PH, Sup, P)
		
		

		return {'V':V, 'H':H, 'S': S}
		
		
	def search_PV(self, P:float, V:float):
		"""
		search for the values at a given pressure (kPa) and specific volume (m3/kg)
		"""
		cursor = self.m_Connection.cursor()

		PL, PH = self._BracketPressure(P)

		Tlow, Hlow, Slow = self._FindProperties(PL, "V", V)
		Tup, Hup, Sup = self._FindProperties(PH, "V", V)

		T = _linearinterp(PL, Tlow, PH, Tup, P)
		H = _linearinterp(PL, Hlow, PH, Hup, P)
		S = _linearinterp(PL, Slow, PH, Sup, P)

		return {'T':T, 'H':H, 'S':S}
		
	
	def search_PH(self, P:float, H:float):
		"""
		search for the values at a given pressure (kPa) and enthalpy (kJ/kg)
		"""
		cursor = self.m_Connection.cursor()

		PL, PH = self._BracketPressure(P)

		Tlow, Vlow, Slow = self._FindProperties(PL, "H", H)
		Tup, Vup, Sup = self._FindProperties(PH, "H", H)

		T = _linearinterp(PL, Tlow, PH, Tup, P)
		V = _linearinterp(PL, Vlow, PH,Vup, P)
		S = _linearinterp(PL, Slow, PH, Sup, P)

		return {'T':T, 'V':V, 'S':S}
	
	
	def search_PS(self, P:float, S:float):
		"""
		search for the values at a given pressure (kPa) and entropy (kJ/kgK)
		"""
		cursor = self.m_Connection.cursor()

		PL, PH = self._BracketPressure(P)

		Tlow, Vlow, Hlow = self._FindProperties(PL, "S",S)
		Tup, Vup, Hup = self._FindProperties(PH, "S", S)

		T = _linearinterp(PL, Tlow, PH, Tup, P)
		V = _linearinterp(PL, Vlow, PH,Vup, P)
		H = _linearinterp(PL, Hlow, PH, Hup, P)

		return {'T':T, 'V':V, 'H':H} 
		
	
	def search(self, P:float, name:str, value:float):
		CapName = name.upper()
		if(CapName == "T"):
			return self.search_PT(P, value)
		
		elif (CapName =="V"):
			return self.search_PV(P, value)
		
		elif(CapName == "H"):
			return self.search_PH(P, value)
		
		elif(CapName == "S"):
			return self.search_PS(P, value)
		
		else:
			raise ValueError("name must be T, V, H or S")
		


class ThermoPhysical():  
	"""
	Thermo-physical properties (T, rho, cp, viscosity, k, Pr) of fluids
	"""
	s_DataBasePath = _parent_path(__file__) / "thermophysical.db"
	
	def __init__(self, FluidName:str ) -> None:
		"""
		FluidName: Name of the fluid
		"""

		self.m_Connection = _sql.connect(self.s_DataBasePath) 
		
		if(FluidName ==""):
			return
		
		self.m_FluidName = FluidName
		cursor = self.m_Connection.cursor()

		"""
		Check if the parameter FluidName is valid
		Note that the columns in the database is configured as
		COLLATE NOCASE, therefore search is not case-sensitive
		"""
		QueryString = "SELECT NAME FROM MAINTABLE where NAME=?"
		rows = cursor.execute(QueryString , (FluidName,)).fetchall()

		#more than 1 name matches
		if(len(rows)>1):
			raise ValueError("More than 1 fluid matched the name:" + FluidName)

		self.m_DBTable = rows[0][0]


	def __del__( self ):
		self.m_Connection.close()
	
	
	def GetFluidNames(self):
		QueryString = "SELECT name FROM MAINTABLE"
		FluidNameList = self.m_Connection.cursor().execute(QueryString , []).fetchall()
		
		return FluidNameList


	def search(self, PropertyName:str, QueryValue:float, Sort:bool = True): 
		return _searchOrderedTable(self.m_Connection, self.m_DBTable, PropertyName, QueryValue, Sort) 



def _GetFieldNames(connection, TableName: str)->list:
		"""
		Given a TableName returns all fieldnames (properties) of the table
		"""
		QueryString = "SELECT name FROM PRAGMA_TABLE_INFO(?)"
		rowList = connection.cursor().execute(QueryString , (TableName,)).fetchall()

		if(len(rowList) == 0):
			raise ValueError("Invalid table name:" + TableName)

		retList = []
		for tupleItem in rowList:
			retList.append(tupleItem[0])

		return retList

	



def _searchOrderedTable(connection, TableName:str, PropertyName:str, QueryValue:float, Sort:bool = True)->dict:
	"""
		TableName: Database table name where uniquely named properties are <br>
		PropertyName: Name of the property corresponding to fieldname in the table <br>
		QueryValue: Value at which properties are sought after <br>
		Sort: Sort the table based on the property name in ascending order 
		(if multiple queries are performed on the same property, set Sort to False for efficiency )
		
		--Table must be in the form of, for example <br>
		P	T	s	vf <br>
		50	20	2	0.2 <br>
		70	25	3	0.8 <br>
		
		if we are after properties at T=22, then returns a dict with keys P, s corresponding to T=22`
	"""
	cursor = connection.cursor() 
	AllFieldNames = _GetFieldNames(connection, TableName)

	#Index of the property in the columns of the table
	ParamIndex = -1
	try:
		"""
			note that we search in capitalized field names with capitalized PropertyName
			For example, user now can enter Pr, pr ... to find properties for Prandtl
			The table must contain only unique fields
		"""
		CapitalizedFiledNames = [s.capitalize() for s in AllFieldNames]
		ParamIndex = CapitalizedFiledNames.index(PropertyName.capitalize())
	except:
		raise ValueError("Valid property names: " + str(AllFieldNames))
	

	#Check if QueryValue is within the bounds of the property
	strQuery="SELECT min( {} ), max( {} ) FROM " + TableName 
	rows = cursor.execute(strQuery.format(PropertyName, PropertyName)).fetchone()
	MinVal, MaxVal = rows[0],  rows[1]
	if(not (MinVal<QueryValue and QueryValue<MaxVal)):
		raise ValueError(PropertyName + " range: [" + str(rows[0]) + " , " + str(rows[1]) + "]")

	_rows = None

	if(Sort):
		strQuery = "SELECT * FROM " + TableName + " ORDER BY "+ PropertyName
		_rows = cursor.execute(strQuery , []).fetchall()
	

	RowIndex = -1
	for i in range(len(_rows)):
		Value = _rows[i][ParamIndex]
		if(Value>=QueryValue):
			RowIndex = i
			break
	
	
	retDict = dict()
	if(RowIndex == 0):
		TupleIndex = -1
		for propName in AllFieldNames:
			TupleIndex += 1
			if(propName == PropertyName):
				continue

			Value = _rows[0][TupleIndex]
			retDict[propName] = Value

		return retDict
			
		

	PropValHigh = _rows[RowIndex][ParamIndex]
	PropValLow = _rows[RowIndex - 1][ParamIndex] 
	TupleIndex = -1

	for propName in AllFieldNames: 
		TupleIndex += 1
		
		if(propName.capitalize() == PropertyName.capitalize()):
			continue 

		ValueLow = _rows[RowIndex - 1][TupleIndex]
		ValueHigh = _rows[RowIndex][TupleIndex]
		Value = _linearinterp(PropValLow, ValueLow, PropValHigh, ValueHigh, QueryValue)
		retDict[propName] = Value
		
	return retDict



import math
import numpy as np

from dataclasses import dataclass

from . import pbinom, pnorm


__all__ = ['test_sign', 'test_sign_Result', 'CI_Result']




@dataclass
class CI_Result:
	pos:int
	prob:float
	CILow:float; CIHigh:float


@dataclass
class test_sign_Result:
	lower:CI_Result
	interpolated:CI_Result
	upper:CI_Result


def _FindCI(xvec, alternative, ConfLevel):
	xvec = np.sort(xvec)

	lower, exact, upper = 0.0, 0.0, 0.0
	lower_found, upper_found = False, False
	lower_pos, upper_ps = 0, 0
	prob = 0.0

	NSample = len(xvec)

	for i in range(NSample):
		_pbinom = pbinom(i, NSample, 0.5)

		if alternative == "two.sided" or alternative == "notequal":
			prob = 1 - 2 * _pbinom

		elif alternative == "greater":
			prob = 1 - _pbinom

		elif alternative == "less":
			prob = _pbinom


		if prob >= ConfLevel:
			upper = prob
			upper_found = True

			upper_pos = i

			if (lower_found): 
				break

		else:
			lower = prob
			lower_found = True

			lower_pos = i

			if (upper_found): 
				break

	CI_a = xvec[lower_pos]
	CI_b = xvec[len(xvec) - lower_pos - 1]

	lowerRes = CI_Result(
		pos = lower_pos,
		prob=lower,
		CILow = min(CI_a, CI_b),
		CIHigh = max(CI_a, CI_b))
	

	CI_a = xvec[upper_pos]
	CI_b = xvec[len(xvec) - upper_pos - 1]

	upperRes = CI_Result(
		pos = upper_pos,
		prob = upper,
		CILow = min(CI_a, CI_b),
		CIHigh = max(CI_a, CI_b))

	return lowerRes, upperRes





def test_sign(x, md, y=None, alternative="two.sided", conflevel=0.95):
	"""
	
	## Return
	p-value and a test_sign_Result class. \n

	## Input
	x: Sample \n
	y: Second sample \n
	md: Median of the population tested by the null hypothesis \n
	alternative: "two.sided", "less", "greater" \n
	conflevel:	Confidence level, [0,1]
	"""

	NORMALAPPROX = 12

	assert conflevel>=0.0 or conflevel <= 1.0, "conflevel must be in range (0, 1)"
	assert isinstance(x, list) or type(x)==np.ndarray, "x must be list/ndarray"

	XX, YY = x, y
	if isinstance(x, list):
		XX = np.asfarray(x)
	
	assert np.issubdtype(XX.dtype, np.number), "x must contain only numbers"
	
	xvec = XX

	if y != None:
		assert isinstance(y, list) or type(y)==np.ndarray, "y must be list/ndarray"
		if isinstance(y, list):
			YY = np.asfarray(y)

		assert np.issubdtype(YY.dtype, np.number), "y must contain only numbers"
		xvec = XX - YY
	
	NGreater = len(np.argwhere(xvec>md))

	EqIndex = np.where(xvec == md)
	NEqual = len(EqIndex)
	xvec = np.delete(xvec, EqIndex)

	pvalue = 0.0

	NSample = len(xvec)

	if NSample < NORMALAPPROX:
		if alternative == "two.sided" or alternative == "notequal":
			n = min(NGreater, NSample - NGreater)

			#B(n, 0.5) is symmetric
			pvalue = 2.0 * pbinom(n, NSample, 0.5)

		elif alternative == "greater":
			pvalue = 1.0 - pbinom(NGreater - 1, NSample, 0.5) #area on the right

		elif alternative == "less":
			pvalue = pbinom(NGreater, NSample, 0.5); #area on the left
	
	else:
		variance = NSample / 4.0
		zvalue = (NGreater - NSample / 2.0) / math.sqrt(variance)

		if alternative == "two.sided" or alternative == "notequal":
			if zvalue <= 0:
				#area on the left + area on the right of positive
				pvalue = pnorm(zvalue, 0.0, 1.0) + (1 - pnorm(-1.0 * zvalue, 0.0, 1.0))
			else:
				#area on the right of positive + area on the left of negative
				pvalue = (1.0 - pnorm(zvalue, 0.0, 1.0)) + pnorm(-zvalue, 0.0, 1.0)
		
		elif alternative == "greater":
			#area on the right
			pvalue = (1.0 - pnorm(zvalue, 0.0, 1.0))

		elif alternative == "less":
			#area on the left
			pvalue = pnorm(zvalue, 0.0, 1.0)

	Lower, Upper = _FindCI(xvec, alternative, conflevel)

	interpolated = CI_Result(pos=-1, prob=conflevel, CILow=0.0, CIHigh=0.0)
	if alternative == "two.sided" or alternative == "notequal":
		x1, x2 = Lower.prob, Upper.prob

		interpolated.CILow = np.interp(conflevel, [x1, x2], [Lower.CILow, Upper.CILow] )
		interpolated.CIHigh = np.interp(conflevel, [x1, x2], [Lower.CIHigh, Upper.CIHigh])
	
	elif alternative == "greater":
		interpolated.CILow = np.interp(conflevel, [Lower.prob, Upper.prob], [Lower.CILow, Upper.CILow])
		interpolated.CIHigh = math.inf
	
	elif alternative == "less":
		interpolated.CILow = -math.inf
		interpolated.CIHigh = np.interp(conflevel, [Lower.prob, Upper.prob], [Lower.CIHigh, Upper.CIHigh])
	
	else:
		raise ValueError("Values for 'alternative': \"two.sided\" or \"notequal\", \"greater\", \"less\"")

	result = test_sign_Result(
		lower= Lower,
		interpolated=interpolated,
		upper = Upper)
	
	return pvalue, result





from .._ctypeslib import core as _ctcore



class PsychrometryResult:
	"""
	<p>Computation results of psychrometry class. </p>

	>>result = psy.compute() <br>
	>>result.P #access to pressure
	"""

	def __init__(self, psyresult):
		self.m_Results = psyresult

	def __str__(self) -> str:
		retStr=""

		if(len(self.m_Results)==0):
			return retStr

		retStr += "P=" + str(round(self.P, 2)) + " kPa" + "\n"

		retStr += "Tdb=" + str(round(self.Tdb, 2)) + " C" + "\n"
		
		retStr += "Twb=" + str(round(self.Twb, 2)) + " C" + "\n"
		
		retStr += "Tdp=" + str(round(self.Tdp, 2)) +  " C" + "\n"
		
		retStr += "H=" + str(round(self.H, 2)) + " kJ/kg da" + "\n"
		
		retStr += "RH=" + str(round(self.RH, 2)) + " %" + "\n"
		
		retStr += "W=" + str(round(self.W, 4)) +  " kg/kg da" + "\n"
		
		retStr += "V=" + str(round(self.V,3)) +  " m3/kg da" + "\n"

		return retStr


            
	@property
	def P(self): 
		"""Pressure in kPa"""
		
		return self.m_Results.get("P")


	@property
	def Tdb(self): 
		"""
		Dry bulb temperature in Celcius
		"""
		return self.m_Results.get("Tdb")

	@property
	def Twb(self): 
		"""
		Wet bulb temperature in Celcius
		"""
		return self.m_Results.get("Twb")

	@property
	def Tdp(self): 
		"""
		Dew point temperature in Celcius
		"""
		return self.m_Results.get("Tdp")

	@property
	def H(self):
		"""
		Enthalpy in kJ/kg da
		""" 
		return self.m_Results.get("H")

	@property
	def RH(self): 
		"""
		Relative Humidity in %
		"""
		return self.m_Results.get("RH")

	@property
	def W(self):
		"""
		Absolute Humidity in kg/kg da
		""" 
		return self.m_Results.get("W")

	@property
	def V(self): 
		"""
		Specific volume in m3/kg da
		"""
		return self.m_Results.get("V")

	@property
	def Pws(self):
		"""
		water vapor pressure at saturation in kPa
		"""
		return self.m_Results.get("Pws")

	@property
	def Pw(self):
		"""
		water vapor pressure in kPa
		"""
		return self.m_Results.get("Pw")

	@property
	def Ws(self):
		"""
		Absolute humidity at saturation kg/kg da
		"""
		return self.m_Results.get("Ws")



def psychrometry(**kwargs):
		
	"""
	Computes thermodynamic properties of humid-air.

	## Input: 
	3 keyword arguments containing only the following keys: \n
	1) P: Pressure (kPa)
	2) Tdb: Dry-bulb temperature (Celcius)
	3) Twb: Wet-bulb temperature (Celcius)
	4) Tdb: Wet-bulb temperature (Celcius)
	5) W: Absolute humidity (kg/kg da)
	6) RH: Relative humidity (%)
	7) H: Enthalpy (kJ/kg da)
	8) V: Specific Volume (m3/ kg da)

	## Example
	p=psychrometry(P=100, Tdb=50, RH=60) \n
	print(p) # prints all properties \n
	p.H # access only to enthalpy

	"""
	return PsychrometryResult(_ctcore.c_eng_psychrometry(kwargs))



import numpy as _np


def rbinom(n:int, size, prob):
	"""
	Draw samples from binomial distribution
	"""
	assert n>0 ,"n>0 expected"
	assert size>0, "size>0 expected"
	assert prob>=0 and prob<=1, "prob in [0, 1] expected"

	return _np.random.binomial(size=n, n=size, p=prob).tolist()


def rchisq(n:int, df):
	"""
	Draw samples from Chi-Square distribution
	"""
	assert n>0 ,"n>0 expected"
	assert df>0, "df>0 expected"

	return _np.random.chisquare(size=n, df = df).tolist()


def rf(n:int, df1, df2):
	"""
	Draw samples from F distribution
	"""
	assert n>0 ,"n>0 expected"
	assert df1>0, "df1>0 expected"
	assert df2>0, "df2>0 expected"

	return _np.random.f(size=n, dfnum=df1, dfden=df2).tolist()


def rnorm(n:int, mean=0.0, sd=1.0):
	"""
	Draw samples from normal distribution
	"""
	assert n>0 ,"n>0 expected"
	assert sd>0, "sd>0 expected"

	return _np.random.normal(size=n, loc=mean, scale=sd).tolist()


def rpois(n:int, mu = 1):
	"""
	Draw samples from Poisson distribution
	"""
	assert n>0 ,"n>0 expected"
	assert mu>0, "mu>0 expected"

	return _np.random.poisson(size=n, lam=mu).tolist()


def rt(n:int, df):
	"""
	Draw samples from standard Student's t distribution
	"""
	assert n>0 ,"n>0 expected"
	assert df>0, "df1>0 expected"

	return _np.random.standard_t(size=n, df=df).tolist()


def runif(n:int, min=0.0, max=1.0):
	"""
	Draw samples from uniform distribution
	"""
	assert n>0 ,"n>0 expected"

	return _np.random.uniform(size=n, low=min, high=max).tolist()


import numpy as np

from dataclasses import dataclass
from . import pf, qf



__all__ = ['test_f', 'test_f_Result']



@dataclass
class test_f_Result:
	Fcritical:float
	df1:int; df2:int
	var1:float; var2:float
	CI_lower:float; CI_upper:float



def test_f(x, y, alternative:str = "two.sided", ratio:float = 1.0, conflevel:float = 0.95)->tuple:
	"""
	Performs F test

	## Return
	p-value and test_f_Result class. \n

	## Input
	x/y: First/second sample, ndarray/list \n
	alternative: "two.sided", "less", "greater" \n
	ratio: Assumed ratio of variances of the samples \n
	conflevel: Confidence level, [0,1] 
	"""
	assert conflevel>=0.0 or conflevel <= 1.0, "conflevel must be in range (0, 1)"
	assert isinstance(x, list) or type(x)==np.ndarray, "x must be list/ndarray"
	assert isinstance(y, list) or type(y)==np.ndarray, "y must be list/ndarray"

	XX, YY = x, y
	if isinstance(x, list):
		XX = np.asfarray(x)

	if isinstance(y, list):
		YY = np.asfarray(y)
	
	assert np.issubdtype(XX.dtype, np.number), "x must contain only numbers"
	assert np.issubdtype(YY.dtype, np.number), "y must contain only numbers"

	alpha = 1 - conflevel
	df1, df2 = len(XX) - 1, len(YY) -1 #degrees of freedoms
	
	var1, var2 = np.var(XX, ddof = 1), np.var(YY, ddof = 1)
	varRatio = var1 / var2

	Fcritical = varRatio / ratio

	pvalue = 0.0

	if alternative == "two.sided" or alternative == "notequal":
		#F distribution is non - symmetric
		a = pf(Fcritical, df1, df2)
		b = 1 - pf(Fcritical, df1, df2)

		c = pf(1 / Fcritical, df1, df2)
		d = 1 - pf(1 / Fcritical, df1, df2)

		pvalue = min(a, b) + min(c, d)

	elif alternative == "greater":
		pvalue = (1 - pf(Fcritical, df1, df2)) #area on the right

	elif alternative == "less":
		pvalue = pf(Fcritical, df1, df2) #area on the left

	else:
		raise ValueError("Values for 'alternative': \"two.sided\" or \"notequal\", \"greater\", \"less\"");


	CI_lower, CI_upper = None, None
	if alternative == "two.sided" or alternative == "notequal":
		CI1 = varRatio * qf(alpha / 2.0, df1, df2)
		CI2 = varRatio * qf(1 - alpha / 2.0, df1, df2)

		CI_lower = min(CI1, CI2)
		CI_upper = max(CI1, CI2)

	elif alternative == "greater":
		CI_lower = varRatio * qf(alpha, df1, df2)

	elif alternative == "less":
		CI_upper = varRatio * qf(1 - alpha, df1, df2)
	
	result = test_f_Result(Fcritical= Fcritical, 
			df1=df1,
			df2=df2,
			var1=var1,
			var2=var2,
			CI_lower=CI_lower,
			CI_upper=CI_upper)

	return pvalue, result


import sys, os

#inserting to 0th position is very important so that search will FIRST match ../scisuit folder
sys.path.insert(0, os.getcwd()) 

import wx
import ctypes

from scisuit.apps import frmPsychrometry, frmFluidProperties

#Important otherwise colors/text of app looks blurry
ctypes.windll.shcore.SetProcessDpiAwareness(True)

app = wx.App()

frm1 = frmPsychrometry()
frm1.Show()

frm2 = frmFluidProperties()
frm2.Show()

app.MainLoop()



from ._ctypeslib import core as _core
import ctypes as _ct
import numbers as _numbers
from numpy.polynomial import polynomial as _Polynomial


__all__ = ['linearinterp', 'lagrange', 'spline', 'expfit', 'logfit', 'logistfit', 'polyfit', 'powfit']



def linearinterp(x1:float, y1:float, x2:float, y2:float, xval:float)->float:
	"""
	Linear interpolation

	## Inputs:
	x1, y1: First point \n
	x2, y2: Second point \n
	xval: x-value in [x1, x2] 

	## Returns:
	y-value in [y1, y2] corresponding to xval
	"""

	if(x1 == x2): 
		return y1 
	
	m,n=0, 0
	m = (y2 - y1) / (x2 - x1)
	n = y2 - m * x2

	return m * xval + n





def lagrange(x, y, val:float)->float:
	"""
	Constructs lagrange polynomial from x,y to compute the given value.

	## Inputs:
	x, y: list/ndarray \n
	val: value in the range of x whose corresponding y value is wish to be known.
	"""
	assert issubclass(val, _numbers.Real)

	return _core.c_fit_lagrange(x, y, _ct.c_double(val))


def spline(x, y)->list:
	"""
	Constructs natural cubic spline polynomials from x, y

	## Inputs:
	x, y: list/ndarray

	## Return:
	Returns a 2D Python list, where each sub-list contains exactly 3 entries: \n
	1) List of coefficients of the cubic polynomial,
	2) Lower bound,
	3) Upper bound.
	"""
	return _core.c_fit_spline(x, y)



def expfit(x, y, intercept=None)->list:
	"""
	Fits x,y to the equation y = a*exp(b*x) \n
	return list is [a, b]

	## Inputs:
	x, y: list/ndarray \n
	intercept: None or float

	## Note:
	If intercept is not provided both a, b are computed. \n
	If intercept is given then a=intercept and b is computed accordingly.
	"""
	if(intercept!=None):
		assert issubclass(intercept, _numbers.Real)

	return _core.c_fit_expfit(x, y, intercept)


def logfit(x, y)->list:
	"""
	Fits x,y to the equation y = a*ln(x) + b \n
	return list is [a, b]

	## Inputs:
	x, y: list/ndarray
	"""

	return _core.c_fit_logfit(x, y)


def logistfit(x, y, limit = None)->list:
	"""
	Fits x,y to the equation y = L / (1 + exp(b0 + b1*x)) \n
	return list is [L, b0, b1] or [b0, b1]

	## Inputs:
	x, y: list/ndarray \n
	limit: None or float
	"""
	if(limit!=None):
		assert issubclass(limit, _numbers.Real)

	return _core.c_fit_logistfit(x, y, limit)


def polyfit(x, y, deg)->tuple:
	"""
	uses NumPy's polynomial fitting (numpy.polynomial.polynomial.polyfit) \n
	returns (coefficients, residuals)

	## Inputs:
	x, y: list/ndarray

	"""
	coeffs, stats = _Polynomial.polyfit(x=x, y=y, deg=deg)
	return coeffs, stats[0]


def powfit(x, y)->list:
	"""
	Fits x,y to the equation y = a*x^n
	returns [a, n]

	## Inputs:
	x, y: list/ndarray
	"""
	return _core.c_fit_powfit(x,y)




import sys, os
import pprint

#inserting to 0th position is very important so that search will FIRST match ../scisuit folder
sys.path.insert(0, os.getcwd()) 


import scisuit.stats as st
import numpy as np

print("*** Normal Distribution ***")

#Compute density
print(st.dnorm(x=0, sd=2.0))

arr1=np.array([-0.5, 0.0, 0.5])
print(st.dnorm(x=arr1))


print("\n *** Binomial Distribution ***")
print(st.dbinom(x=2, size=5, prob=0.3))

arr2 = np.array([7, 8, 9])
print(st.dbinom(x=arr2, size=9, prob=0.94))


print("\n *** Chi-Square Distribution ***")
print(st.dchisq(x=2, df=5))

arr3 = np.array([0, 1, 2])
print(st.dchisq(x=arr3, df=5))


"""
from scisuit.stats import test_t

treat = [24, 43, 58, 71, 43, 49, 61, 44, 67, 49]
cont = [42, 43, 55, 54, 20, 85, 33, 41, 19, 60, 53, 42]

pval, tbl=test_t(x=treat, y=cont, varequal=False)

print("p-value=" + str(pval))
pprint.pprint(tbl)
"""

"""
import numpy as np
from scisuit.stats import aov

A = np.array([16, 11, 20, 21, 14, 7])
B = np.array([21, 12, 14, 17, 13, 17])
C = np.array([37, 32, 12, 25, 39, 41])
D = np.array([45, 59, 48, 46, 38, 47])

#perform 1-way ANOVA
cls = aov(A, B, C, D)

#p-value and extra info
pval, info = cls.compute()

print("p-value=" + str(pval))
pprint.pprint(info)

"""


import wx
import os
import pkgutil
from .path import pyhomepath


__all__=['pkg_installed', 'assert_pkg']


def pkg_installed(name:str)->bool:
	"""
	checks if the package with given name installed \n

	For example, wxPython's package name is wx but pip installation
	requires wxPython. Here the name parameter is wx.
	"""
	x = pkgutil.iter_modules()
	for i in x:
		if i.ispkg==True and i.name == name:
			return True
	
	return False


def assert_pkg(name:str, pip:str)->bool:
	"""
	## Input: 
	Name: package name, (wx) \n
	pip: pip install name (wxPython) 

	## Return: 
	if package is already installed returns True \n

	If missing, prompts the user (wx.MessageBox) to install the package, \n
	if user clicks No, returns False, \n
	if user clicks Yes, installation process begins and function returns True
	"""

	#package already installed
	if(pkg_installed(name)):
		return True
	
	PyHome = pyhomepath()
	PyExe = PyHome + os.sep + "python.exe"
	Cmd = "\"" + PyExe + "\""
	Cmd += " -m pip install " + pip

	Msg = name + " is missing. Wanna install? \n \n"
	Msg += "Choosing Yes will launch the terminal and installation process using the following command: \n \n"
	
	Msg += Cmd + "\n \n"

	Msg += "If you choose No, you might have to manually install the package to run the requiring app."

	YesNo = wx.MessageBox(Msg, "Install " + name + "?", wx.YES_NO)
	if (YesNo == wx.NO):
		return False

	wx.Shell(Cmd)

	return True



import ctypes as _ct
from .._ctypeslib import core as _core
from dataclasses import dataclass as _dataclass    


def dbinom(x, size:int, prob:float):
	return _core.c_stat_dbinom(x, _ct.c_int(size), _ct.c_double(prob))


def pbinom(q, size:int, prob:float):
	return _core.c_stat_pbinom(q, _ct.c_int(size), _ct.c_double(prob))

def qbinom(p, size:int, prob:float):
	return _core.c_stat_qbinom(p, _ct.c_int(size), _ct.c_double(prob))


def df(x, df1:int, df2:int):
	return _core.c_stat_df(x, _ct.c_int(df1), _ct.c_int(df2))

def pf(q, df1:int, df2:int):
	return _core.c_stat_pf(q, _ct.c_int(df1), _ct.c_int(df2))

def qf(p, df1:int, df2:int):
	return _core.c_stat_qf(p, _ct.c_int(df1), _ct.c_int(df2))

#----------------
def dnorm(x, mean=0.0, sd=1.0):
	return _core.c_stat_dnorm(x, _ct.c_double(mean), _ct.c_double(sd))

def pnorm(q, mean=0.0, sd=1.0):
	return _core.c_stat_pnorm(q, _ct.c_double(mean), _ct.c_double(sd))

def qnorm(p, mean=0.0, sd=1.0):
	return _core.c_stat_qnorm(p, _ct.c_double(mean), _ct.c_double(sd))


#------------------
def dpois(x, mu:float):
	return _core.c_stat_dpois(x, _ct.c_double(mu))

def ppois(q, mu:float):
	return _core.c_stat_ppois(q, _ct.c_double(mu))

def qpois(p, mu:float):
	return _core.c_stat_qpois(p, _ct.c_double(mu))


#-----------------
def dt(x, df:int):
	return _core.c_stat_dt(x, _ct.c_int(df))

def pt(q, df:int):
	return _core.c_stat_pt(q, _ct.c_int(df))

def qt(p, df:int):
	return _core.c_stat_qt(p, _ct.c_int(df))


#----------------------------
def dchisq(x, df:int):
	return _core.c_stat_dchisq(x, _ct.c_int(df))

def pchisq(q, df:int):
	return _core.c_stat_pchisq(q, _ct.c_int(df))

def qchisq(p, df:int):
	return _core.c_stat_qchisq(p, _ct.c_int(df))


#-------------------
def dunif(x, min=0.0, max=1.0):
	return _core.c_stat_dunif(x, _ct.c_double(min), _ct.c_double(max))

def punif(q, min=0.0, max=1.0):
	return _core.c_stat_punif(q, _ct.c_double(min), _ct.c_double(max))

def qunif(p, min=0.0, max=1.0):
	return _core.c_stat_qunif(p, _ct.c_double(min), _ct.c_double(max))


#----------------
def dsignrank(x, n:int):
	return _core.c_stat_dsignrank(x, _ct.c_int(n))

def psignrank(q, n:int):
	return _core.c_stat_psignrank(q, _ct.c_int(n))

def qsignrank(p, n:int):
	return _core.c_stat_qsignrank(p, _ct.c_int(n))


#----------------------
def moveavg(x, y, period=2):
	return _core.c_stat_moveavg(x, y, _ct.c_int(period))

def rolling(x, y, period=2):
	return _core.c_stat_rolling(x, y, _ct.c_int(period))

#----------------------

@_dataclass
class TestNormRes:
	pval:float
	A2:float

def test_norm_ad(x):
	pval, A2 = _core.c_stat_test_norm_ad(x)
	return TestNormRes(pval, A2)





from .aov import aov
from .aov2 import aov2, aov2_results
from .kurt import kurt
from .linregress import linregress
from .rolling import rolling
from .mode import mode
from .sample import sample
from .skew import skew
from .test_z import test_z, test_z_Result
from .test_f import test_f, test_f_Result
from .test_sign import test_sign, test_sign_Result
from .test_t import test_t, test_t1_result, test_t2_result, test_tpaired_result

from .dist_random import rbinom, rchisq, rf, rnorm, rpois, rt, runif







import os as _os
import wx as _wx
from pathlib import Path as _Path

def makeicon(path:_Path)->_wx.Icon:
	"""
	path: image's full path 
	"""

	if(not path.is_absolute()):
		raise ValueError(path + " is relative path, full path expected.")
		
	if(not path.exists()):
		raise ValueError("Invalid path: " + path)

	icon = _wx.Icon()
	image = _wx.Image()
	image.LoadFile(str(path))
	bmp=image.ConvertToBitmap()
	icon.CopyFromBitmap(bmp)

	return icon





import sys, os
import pprint

#inserting to 0th position is very important so that search will FIRST match ../scisuit folder
sys.path.insert(0, os.getcwd()) 


from scisuit.eng import SaturatedRefrigerant

#saturated properties of water
fl = SaturatedRefrigerant("water")

#search properties for T=10°C
result = fl.search("T", 10)

#print dictionary containing properties
pprint.pprint(result)

print("\n ----------- \n")

#print enthalpy
print("hg=" + str(result["hg"]))


import os as _os
import sys as _sys
from pathlib import Path as _Path



def parent_path(path:str, level = 0)->_Path:
	"""
	This is used so many times that a convenience function is deemed appropriate.

	## Input: 
	path: Relative or full path 
	
	## Example:
	if path is C:\\a\\b\\c.py  \n
	level=0 => C:\\a\\b \n
	level=1 => C:\\a
	
	"""
	pt = _Path(path)

	if(pt.is_absolute() == False):
		pt = pt.absolute()

	return pt.parents[level]



def pyhomepath()->str:
	"""
	returns the Python Home Path
	"""
	return _sys.exec_prefix



import wx

from ..util import parent_path as _parent_path
from ..wxpy import NumTextCtrl, makeicon
from ..eng.fluids_dbase import Refrigerant, SaturatedRefrigerant, SuperHeatedRefrigerant, ThermoPhysical


class pnlRefrigerantSaturated ( wx.Panel ):

	def __init__( self, parent):
		wx.Panel.__init__ ( self, parent )

		self.m_txtBGChanged  = None
		self.m_FluidType = None
		self.m_SelectedProperty = None
		self.m_Parent = parent.GetParent()

		self.m_stFluidType = wx.StaticText( self, wx.ID_ANY, u"Type: ")
		self.m_stFluidType.Wrap( -1 )

		self.m_choiceFluidType = wx.Choice( self, choices = [])
		self.m_choiceFluidType.SetSelection( 0 )

		self.m_radioT = wx.RadioButton( self, label = u"T (°C)")
		self.m_txtT = NumTextCtrl( self)
		self.m_radioP = wx.RadioButton( self, label = u"P (kPa)")
		self.m_txtP = NumTextCtrl( self)
		self.m_radioVf = wx.RadioButton( self, label = u"vf (m\u00B3/kg)")
		self.m_txtVf = NumTextCtrl( self)
		self.m_radioVg = wx.RadioButton( self, label = u"vg (m\u00B3/kg)")
		self.m_txtVg = NumTextCtrl( self)
		self.m_radioHf = wx.RadioButton( self, label = u"hf (kJ/kg)")
		self.m_txtHf = NumTextCtrl( self)
		self.m_radioHg = wx.RadioButton( self, label = u"hg (kJ/kg)")
		self.m_txtHg = NumTextCtrl( self)
		self.m_radioSf = wx.RadioButton( self, label = u"sf (kJ/kg\u00B7K)")
		self.m_txtSf = NumTextCtrl( self)
		self.m_radioSg = wx.RadioButton( self, label = u"sg (kJ/kg\u00B7K)")
		self.m_txtSg = NumTextCtrl( self)

		sizerFluidType = wx.BoxSizer( wx.HORIZONTAL )
		sizerFluidType.Add( self.m_stFluidType, 0, wx.ALL, 5 )
		sizerFluidType.Add( self.m_choiceFluidType, 1, wx.ALL, 5 )

		fgSzr = wx.FlexGridSizer( 0, 2, 0, 0 )
		fgSzr.AddGrowableCol( 1 )
		fgSzr.SetFlexibleDirection( wx.BOTH )
		fgSzr.SetNonFlexibleGrowMode( wx.FLEX_GROWMODE_SPECIFIED )
		fgSzr.Add( self.m_radioT, 0, wx.ALL, 5 )
		fgSzr.Add( self.m_txtT, 0, wx.ALL|wx.EXPAND, 5 )
		fgSzr.Add( self.m_radioP, 0, wx.ALL, 5 )
		fgSzr.Add( self.m_txtP, 0, wx.ALL|wx.EXPAND, 5 )
		fgSzr.Add( self.m_radioVf, 0, wx.ALL, 5 )
		fgSzr.Add( self.m_txtVf, 0, wx.ALL|wx.EXPAND, 5 )
		fgSzr.Add( self.m_radioVg, 0, wx.ALL, 5 )
		fgSzr.Add( self.m_txtVg, 0, wx.ALL|wx.EXPAND, 5 )
		fgSzr.Add( self.m_radioHf, 0, wx.ALL, 5 )
		fgSzr.Add( self.m_txtHf, 0, wx.ALL|wx.EXPAND, 5 )
		fgSzr.Add( self.m_radioHg, 0, wx.ALL, 5 )
		fgSzr.Add( self.m_txtHg, 0, wx.ALL|wx.EXPAND, 5 )
		fgSzr.Add( self.m_radioSf, 0, wx.ALL, 5 )
		fgSzr.Add( self.m_txtSf, 0, wx.ALL|wx.EXPAND, 5 )
		fgSzr.Add( self.m_radioSg, 0, wx.ALL, 5 )
		fgSzr.Add( self.m_txtSg, 0, wx.ALL|wx.EXPAND, 5 )

		self.m_btnCompute = wx.Button( self, wx.ID_ANY, u"Compute")

		mainSizer = wx.BoxSizer( wx.VERTICAL )
		mainSizer.Add( sizerFluidType, 1, wx.EXPAND, 5 )
		mainSizer.Add( ( 0, 10), 1, wx.EXPAND, 5 )
		mainSizer.Add( fgSzr, 0, wx.EXPAND, 5 )
		mainSizer.Add( ( 0, 10), 1, wx.EXPAND, 5 )
		mainSizer.Add( self.m_btnCompute, 0, wx.ALL|wx.ALIGN_CENTER_HORIZONTAL, 5 )

		self.SetSizerAndFit( mainSizer )
		self.Layout()

		self.m_CtrlList = [[self.m_radioT, self.m_txtT, "T"], [self.m_radioP, self.m_txtP, "P"], 
            [self.m_radioVf, self.m_txtVf, "vf"], [self.m_radioVg, self.m_txtVg, "vg"],
            [self.m_radioHf, self.m_txtHf, "hf"], [self.m_radioHg, self.m_txtHg, "hg"],
            [self.m_radioSf, self.m_txtSf, "sf"], [self.m_radioSg, self.m_txtSg, "sg"]]


		self.Bind( wx.EVT_INIT_DIALOG, self.__OnInitDialog )
		self.Bind( wx.EVT_RADIOBUTTON, self.__OnRadioButton )

		self.m_choiceFluidType.Bind( wx.EVT_CHOICE, self.__OnChoice )	
		self.m_btnCompute.Bind( wx.EVT_BUTTON, self.__OnComputeBtn )


	def __del__( self ):
		pass


	def __ChangeBGColor(self, txtCtrl):
		BGColor = wx.Colour(144,238,144)
		if(self.m_txtBGChanged != None):
			self.m_txtBGChanged.SetBackgroundColour(wx.Colour(255, 255, 255)) 
			self.m_txtBGChanged.Refresh()
		
		txtCtrl.SetBackgroundColour(BGColor)
		txtCtrl.Refresh()
		self.m_txtBGChanged = txtCtrl


	def __OnInitDialog( self, event ):
		refrigerant = Refrigerant()
		self.m_FluidList = refrigerant.GetFluidNames()
	
		for entry in self.m_FluidList:
			Name = str(entry[0])
			Alternative = str(entry[1])
			if(len(Alternative)>7):
				Alternative = Alternative[0:int(len(Alternative)/2)] + "..." + Alternative[-int(len(Alternative)/4):]
			
			self.m_choiceFluidType.Append(Name +", "+ Alternative)
		event.Skip()



	def __OnChoice( self, event ):
		sel = event.GetSelection()
		self.m_FluidType = self.m_FluidList[sel][0] #name
		event.Skip()



	def __OnRadioButton( self, event ):
		for lst in self.m_CtrlList:
			if(event.GetEventObject() == lst[0]):
				self.__ChangeBGColor(lst[1])
				self.m_SelectedProperty = lst[2]



	def __OnComputeBtn( self, event ):
		try:
			assert self.m_FluidType != None, "Fluid type must be selected"
			assert self.m_SelectedProperty != None, "A property must be selected"
			assert self.m_txtBGChanged.GetValue() != "", "A value must be entered for " + self.m_SelectedProperty
			
			fl = SaturatedRefrigerant(self.m_FluidType ) 
			result = fl.search(self.m_SelectedProperty, float(self.m_txtBGChanged.GetValue()))	
				
			for lst in self.m_CtrlList:
				if lst[1] == self.m_txtBGChanged: continue

				Value = result.get(lst[2])
				Digits = self.m_Parent.GetDigits() 
				lst[1].SetValue(str(round(Value, Digits)) if Digits!=None else str(Value))
		
		except Exception as e:
			wx.MessageBox(str(e))
	
	


	def Export(self):
		s = str(self.m_SelectedProperty) + "=" + str(self.m_txtBGChanged.GetValue()) + "\n"
		
		for lst in self.m_CtrlList: 
			if lst[1] == self.m_txtBGChanged:
				continue
			s += str(lst[2]) + "=" + str(lst[1].GetValue()) + "\n"
		
		if wx.TheClipboard.Open():
			wx.TheClipboard.SetData(wx.TextDataObject(s))
			wx.TheClipboard.Close()
			wx.TheClipboard.Flush()





class pnlRefrigerantSuperheated ( wx.Panel ):

	def __init__( self, parent):
		wx.Panel.__init__ ( self, parent)
		
		self.m_FluidType = None
		self.m_Parent = parent.GetParent()
		self.m_NCheckedBoxes = 1 # pressure already checked

		self.m_staticFluidType = wx.StaticText( self, wx.ID_ANY, u"Type: ")
		self.m_staticFluidType.Wrap( -1 )
		self.m_choiceFluidType = wx.Choice( self, choices = [] )
		self.m_choiceFluidType.SetSelection( 0 )

		self.m_chkT = wx.CheckBox( self, wx.ID_ANY, u"T (°C)" )
		self.m_txtT = NumTextCtrl( self)
		self.m_chkP = wx.CheckBox( self, wx.ID_ANY, u"P (kPa)")
		self.m_chkP.SetValue(True)
		self.m_chkP.Enable( False )
		self.m_txtP = NumTextCtrl( self)
		self.m_txtP.SetBackgroundColour(wx.Colour(144,238,144))
		self.m_chkV = wx.CheckBox( self, wx.ID_ANY, u"v (m\u00B3/kg)")
		self.m_txtV = NumTextCtrl( self)
		self.m_chkH = wx.CheckBox( self, wx.ID_ANY, u"h (kJ/kg)")
		self.m_txtH = NumTextCtrl( self)
		self.m_chkS = wx.CheckBox( self, wx.ID_ANY, u"s (kJ/kg\u00B7K)")
		self.m_txtS = NumTextCtrl( self)

		sizerFluidType = wx.BoxSizer( wx.HORIZONTAL )
		sizerFluidType.Add( self.m_staticFluidType, 0, wx.ALL, 5 )
		sizerFluidType.Add( self.m_choiceFluidType, 1, wx.ALL, 5 )

		fgSzr= wx.FlexGridSizer( 0, 2, 0, 0 )
		fgSzr.AddGrowableCol( 1 )
		fgSzr.SetFlexibleDirection( wx.BOTH )
		fgSzr.SetNonFlexibleGrowMode( wx.FLEX_GROWMODE_SPECIFIED )
		fgSzr.Add( self.m_chkT, 0, wx.ALL, 5 )
		fgSzr.Add( self.m_txtT, 0, wx.ALL|wx.EXPAND, 5 )
		fgSzr.Add( self.m_chkP, 0, wx.ALL, 5 )
		fgSzr.Add( self.m_txtP, 0, wx.ALL|wx.EXPAND, 5 )
		fgSzr.Add( self.m_chkV, 0, wx.ALL, 5 )
		fgSzr.Add( self.m_txtV, 0, wx.ALL|wx.EXPAND, 5 )
		fgSzr.Add( self.m_chkH, 0, wx.ALL, 5 )
		fgSzr.Add( self.m_txtH, 0, wx.ALL|wx.EXPAND, 5 )
		fgSzr.Add( self.m_chkS, 0, wx.ALL, 5 )
		fgSzr.Add( self.m_txtS, 0, wx.ALL|wx.EXPAND, 5 )

		self.m_btnCompute = wx.Button( self, wx.ID_ANY, u"Compute" )

		mainSizer = wx.BoxSizer( wx.VERTICAL )
		mainSizer.Add( sizerFluidType, 1, wx.EXPAND, 5 )
		mainSizer.Add( ( 0, 10), 1, wx.EXPAND, 5 )
		mainSizer.Add( fgSzr, 0, wx.EXPAND, 5 )
		mainSizer.Add( ( 0, 10), 1, wx.EXPAND, 5 )
		mainSizer.Add( self.m_btnCompute, 0, wx.ALL|wx.ALIGN_CENTER_HORIZONTAL, 5 )


		self.SetSizerAndFit( mainSizer )
		self.Layout()
		
		self.m_CtrlList = [
		[self.m_chkT, self.m_txtT,  "T"], 
		[self.m_chkV, self.m_txtV, "v"],
		[self.m_chkH, self.m_txtH, "h"],
		[self.m_chkS, self.m_txtS, "s"]]

		self.Bind( wx.EVT_INIT_DIALOG, self.__OnInitDialog )
		self.m_choiceFluidType.Bind( wx.EVT_CHOICE, self.FluidType_OnChoice )
		self.m_chkT.Bind( wx.EVT_CHECKBOX, self.chkT_OnCheckBox )
		self.m_chkV.Bind( wx.EVT_CHECKBOX, self.chkV_OnCheckBox )
		self.m_chkH.Bind( wx.EVT_CHECKBOX, self.chkH_OnCheckBox )
		self.m_chkS.Bind( wx.EVT_CHECKBOX, self.chkS_OnCheckBox )
		self.m_btnCompute.Bind( wx.EVT_BUTTON, self.btnCompute_OnButtonClick )

	def __del__( self ):
		pass


	def EnableDisable(self, chkCtrl):
		BGColor = wx.Colour(144,238,144)
		White = wx.Colour(255, 255, 255)
		LRed = wx.Colour(255,204,204) #light red
		self.m_NCheckedBoxes += (1 if chkCtrl.GetValue() else -1)
			
		for Ctrls in self.m_CtrlList:
			cond = self.m_NCheckedBoxes == 2
			if(Ctrls[0] == chkCtrl):
				Ctrls[1].SetBackgroundColour(BGColor if cond else White) 
				Ctrls[1].Refresh()
				continue

			Ctrls[1].SetBackgroundColour(LRed if cond else White) 
			Ctrls[0].Enable(not cond)
			Ctrls[1].Refresh()

	
	def __OnInitDialog( self, event ):
		refrigerant = Refrigerant()
		self.m_FluidList = refrigerant.GetFluidNames()
	
		for entry in self.m_FluidList:
			Name = str(entry[0])
			Alternative = str(entry[1])
			if(len(Alternative)>7):
				Alternative = Alternative[0:int(len(Alternative)/2)] + "..." + Alternative[-int(len(Alternative)/4):]
			
			self.m_choiceFluidType.Append(Name +", "+ Alternative)
		event.Skip()
		

	def FluidType_OnChoice( self, event ):
		sel = event.GetSelection()
		self.m_FluidType = self.m_FluidList[sel][0] #name
		event.Skip()

	def chkT_OnCheckBox( self, event ):
		self.EnableDisable(self.m_chkT)
		event.Skip()

	def chkV_OnCheckBox( self, event ):
		self.EnableDisable(self.m_chkV)
		event.Skip()

	def chkH_OnCheckBox( self, event ):
		self.EnableDisable(self.m_chkH)
		event.Skip()

	def chkS_OnCheckBox( self, event ):
		self.EnableDisable(self.m_chkS)
		event.Skip()

	def btnCompute_OnButtonClick( self, event ):
		if self.m_FluidType == None:
			wx.MessageBox("Fluid type must be selected")
			return
		
		if self.m_NCheckedBoxes ==1:
			wx.MessageBox("Exactly two properties must be selected")
			return
		
		if(self.m_txtP.GetValue() == wx.EmptyString):
			wx.MessageBox("A value must be entered for pressure")
			return
		
		SelectedProp:str = ""
		InputVal = ""
		for lst in self.m_CtrlList:
			if lst[0].GetValue() == True:
				SelectedProp = lst[2].upper()
				InputVal = lst[1].GetValue()
		
		if(InputVal == ""):
			wx.MessageBox("A value must be entered for " + SelectedProp)
			return
				
		fl = SuperHeatedRefrigerant(self.m_FluidType ) 
		
		result = dict()
		try:
			result = fl.search(float(self.m_txtP.GetValue()), SelectedProp, float(InputVal))
		except Exception as e:
			wx.MessageBox(str(e))
			return
		
		
		Digits = self.m_Parent.GetDigits()
		
		for Ctrl in  self.m_CtrlList:
			Value = result.get(Ctrl[2].capitalize())
			if(Value == None):
				continue
				
			if(Digits != None):
				Ctrl[1].SetValue(str(round(Value, Digits )))
			else:
				Ctrl[1].SetValue(str(Value))
		
		event.Skip()
	

	def Export(self):
		s = "P=" + str(self.m_txtP.GetValue()) + "\n"

		for lst in self.m_CtrlList: 
			if lst[0].GetValue():
				s += str(lst[2]) + "=" + str(lst[1].GetValue()) + "\n"
				break
		
		for lst in self.m_CtrlList: 
			if lst[0].GetValue():
				continue
			s += str(lst[2]) + "=" + str(lst[1].GetValue()) + "\n"
		
		if wx.TheClipboard.Open():
			wx.TheClipboard.SetData(wx.TextDataObject(s))
			wx.TheClipboard.Close()
			wx.TheClipboard.Flush()





class pnlThermoPhysical ( wx.Panel ):

	def __init__( self, parent):
		wx.Panel.__init__ ( self, parent)
		
		self.m_txtBGChanged  = None
		self.m_FluidType = None
		self.m_SelectedProperty = None
		self.m_Parent = parent.GetParent()

		mainSizer = wx.BoxSizer( wx.VERTICAL )

		sizerFluidType = wx.BoxSizer( wx.HORIZONTAL )

		self.m_staticFluidType = wx.StaticText( self, wx.ID_ANY, u"Type: ")
		self.m_staticFluidType.Wrap( -1 )

		sizerFluidType.Add( self.m_staticFluidType, 0, wx.ALL, 5 )

		m_choiceFluidTypeChoices = []
		self.m_choiceFluidType = wx.Choice( self, wx.ID_ANY, wx.DefaultPosition, wx.DefaultSize, m_choiceFluidTypeChoices, 0 )
		self.m_choiceFluidType.SetSelection( 0 )
		sizerFluidType.Add( self.m_choiceFluidType, 1, wx.ALL, 5 )


		mainSizer.Add( sizerFluidType, 1, wx.EXPAND, 5 )


		mainSizer.Add( ( 0, 10), 1, wx.EXPAND, 5 )

		fgSizerLeft = wx.FlexGridSizer( 0, 2, 0, 0 )
		fgSizerLeft.AddGrowableCol( 1 )
		fgSizerLeft.SetFlexibleDirection( wx.BOTH )
		fgSizerLeft.SetNonFlexibleGrowMode( wx.FLEX_GROWMODE_SPECIFIED )

		self.m_radioT = wx.RadioButton( self, wx.ID_ANY, u"T (°C)")
		fgSizerLeft.Add( self.m_radioT, 0, wx.ALL, 5 )

		self.m_txtT = NumTextCtrl( self)
		fgSizerLeft.Add( self.m_txtT, 0, wx.ALL|wx.EXPAND, 5 )

		self.m_radioRho = wx.RadioButton( self, wx.ID_ANY, u"Density (kg/m\u00B3)")
		fgSizerLeft.Add( self.m_radioRho, 0, wx.ALL, 5 )

		self.m_txtRho = NumTextCtrl( self)
		fgSizerLeft.Add( self.m_txtRho, 0, wx.ALL|wx.EXPAND, 5 )

		self.m_radioCp = wx.RadioButton( self, wx.ID_ANY, u"Cp (kJ/kgK)")
		fgSizerLeft.Add( self.m_radioCp, 0, wx.ALL, 5 )

		self.m_txtCp = NumTextCtrl( self)
		fgSizerLeft.Add( self.m_txtCp, 0, wx.ALL|wx.EXPAND, 5 )

		self.m_radioK = wx.RadioButton( self, wx.ID_ANY, u"k (W/mK)")
		fgSizerLeft.Add( self.m_radioK, 0, wx.ALL, 5 )

		self.m_txtK = NumTextCtrl( self)
		fgSizerLeft.Add( self.m_txtK, 0, wx.ALL|wx.EXPAND, 5 )

		self.m_radioMu = wx.RadioButton( self, wx.ID_ANY, u"Viscosity (Pa s)")
		fgSizerLeft.Add( self.m_radioMu, 0, wx.ALL, 5 )

		self.m_txtMu = NumTextCtrl( self)
		fgSizerLeft.Add( self.m_txtMu, 0, wx.ALL|wx.EXPAND, 5 )

		self.m_radioPr = wx.RadioButton( self, wx.ID_ANY, u"Pr")
		fgSizerLeft.Add( self.m_radioPr, 0, wx.ALL, 5 )

		self.m_txtPr = NumTextCtrl( self)
		fgSizerLeft.Add( self.m_txtPr, 0, wx.ALL|wx.EXPAND, 5 )


		mainSizer.Add( fgSizerLeft, 0, wx.EXPAND, 5 )
		mainSizer.Add( ( 0, 10), 1, wx.EXPAND, 5 )

		self.m_btnCompute = wx.Button( self, wx.ID_ANY, u"Compute")
		mainSizer.Add( self.m_btnCompute, 0, wx.ALL|wx.ALIGN_CENTER_HORIZONTAL, 5 )


		self.SetSizerAndFit( mainSizer )
		self.Layout()
		
		self.m_CtrlList = [[self.m_txtT, "T"], [self.m_txtRho, "rho"], 
            [self.m_txtCp, "cp"], [self.m_txtMu, "mu"],
            [self.m_txtK, "k"], [self.m_txtPr, "Pr"]]

		self.Bind( wx.EVT_INIT_DIALOG, self.OnInitDialog )
		self.m_choiceFluidType.Bind( wx.EVT_CHOICE, self.FluidType_OnChoice )
		self.m_radioT.Bind( wx.EVT_RADIOBUTTON, self.radioT_OnRadioButton )
		self.m_radioRho.Bind( wx.EVT_RADIOBUTTON, self.radioRho_OnRadioButton )
		self.m_radioCp.Bind( wx.EVT_RADIOBUTTON, self.radioCp_OnRadioButton )
		self.m_radioK.Bind( wx.EVT_RADIOBUTTON, self.radioK_OnRadioButton )
		self.m_radioMu.Bind( wx.EVT_RADIOBUTTON, self.radioMu_OnRadioButton )
		self.m_radioPr.Bind( wx.EVT_RADIOBUTTON, self.radioPr_OnRadioButton )
		self.m_btnCompute.Bind( wx.EVT_BUTTON, self.btnCompute_OnButtonClick )

	def __del__( self ):
		pass

	def ChangeBGColor(self, txtCtrl):
		BGColor = wx.Colour(144,238,144)
		if(self.m_txtBGChanged != None):
			self.m_txtBGChanged.SetBackgroundColour(wx.Colour(255, 255, 255)) 
			self.m_txtBGChanged.Refresh()
		
		txtCtrl.SetBackgroundColour(BGColor)
		txtCtrl.Refresh()
		self.m_txtBGChanged = txtCtrl


	def OnInitDialog( self, event ):
		fl = ThermoPhysical("")
		self.m_FluidList = fl.GetFluidNames()
	
		for entry in self.m_FluidList:
			self.m_choiceFluidType.Append(str(entry[0]))
		event.Skip()
	
	
	
	def FluidType_OnChoice( self, event ):
		sel = event.GetSelection()
		self.m_FluidType = self.m_FluidList[sel][0] 
		event.Skip()



	def radioT_OnRadioButton( self, event ):
		self.ChangeBGColor(self.m_txtT)
		self.m_SelectedProperty = "T"
		event.Skip()

	def radioRho_OnRadioButton( self, event ):
		self.ChangeBGColor(self.m_txtRho)
		self.m_SelectedProperty = "rho"
		event.Skip()

	def radioCp_OnRadioButton( self, event ):
		self.ChangeBGColor(self.m_txtCp)
		self.m_SelectedProperty = "cp"
		event.Skip()

	def radioK_OnRadioButton( self, event ):
		self.ChangeBGColor(self.m_txtK)
		self.m_SelectedProperty = "k"
		event.Skip()

	def radioMu_OnRadioButton( self, event ):
		self.ChangeBGColor(self.m_txtMu)
		self.m_SelectedProperty = "mu"
		event.Skip()

	def radioPr_OnRadioButton( self, event ):
		self.ChangeBGColor(self.m_txtPr)
		self.m_SelectedProperty = "pr"
		event.Skip()

	def btnCompute_OnButtonClick( self, event ):
		if self.m_FluidType == None:
			wx.MessageBox("Fluid type must be selected")
			return
		
		if self.m_SelectedProperty == None:
			wx.MessageBox("A property must be selected")
			return
		
		
		if self.m_txtBGChanged.GetValue() == "":
			wx.MessageBox("A value must be entered for " + self.m_SelectedProperty)
			return
		
		fl = ThermoPhysical(self.m_FluidType ) 
		result = dict()
		try:
			result = fl.search(self.m_SelectedProperty, float(self.m_txtBGChanged.GetValue()))
		except Exception as e:
			wx.MessageBox(str(e))
			return
            
		for lst in self.m_CtrlList:
			if lst[0] == self.m_txtBGChanged:
				continue
			Value = result.get(lst[1])
			Digits = self.m_Parent.GetDigits() 
			if(Digits != None):
				lst[0].SetValue(str(round(Value, Digits)))
			else:
				lst[0].SetValue(str(Value)) 

		event.Skip()


	def Export(self):
		s = str(self.m_SelectedProperty) + "=" + str(self.m_txtBGChanged.GetValue()) + "\n"
		
		for lst in self.m_CtrlList: 
			if lst[0] == self.m_txtBGChanged:
				continue
			Name = lst[1]
			Value = lst[0].GetValue()
			s += str(Name) + "=" + str(Value) + "\n"
		
		if wx.TheClipboard.Open():
			wx.TheClipboard.SetData(wx.TextDataObject(s))
			wx.TheClipboard.Close()
			wx.TheClipboard.Flush()



class frmFluidProperties ( wx.Frame ):

	def __init__( self, parent=None, FileMenu = None ):
		wx.Frame.__init__ ( self, parent, id = wx.ID_ANY, title = u"Properties of Fluids" )
		
		self.m_Digits = None
		
		ParentPath = _parent_path(__file__)
		IconPath = ParentPath / "icons" / "fluid.bmp"
		
		self.SetIcon(makeicon(IconPath))

		self.SetSizeHints( wx.DefaultSize, wx.DefaultSize )

		mainSizer = wx.BoxSizer( wx.VERTICAL )

		self.m_notebook = wx.Notebook( self, wx.ID_ANY)
		self.m_pnlSaturated = pnlRefrigerantSaturated( self.m_notebook )
		self.m_pnlSaturated.InitDialog()
		
		self.m_pnlSuperheated = pnlRefrigerantSuperheated( self.m_notebook)
		self.m_pnlSuperheated.InitDialog()
		
		self.m_pnlThermal = pnlThermoPhysical( self.m_notebook)
		self.m_pnlThermal.InitDialog()
		
		self.m_notebook.AddPage( self.m_pnlSaturated, u"Saturated", True )
		self.m_notebook.AddPage( self.m_pnlSuperheated, u"Superheated", False )
		self.m_notebook.AddPage( self.m_pnlThermal, u"Thermo-physical", False )

		mainSizer.Add( self.m_notebook, 1, wx.EXPAND |wx.ALL, 5 )


		self.SetSizerAndFit( mainSizer )
		self.Layout()
		
		self.m_menubar = wx.MenuBar( 0 )

		if FileMenu == None:
			self.m_FileMenu = wx.Menu()
			self.m_ItemExport = wx.MenuItem( self.m_FileMenu, wx.ID_ANY, u"Copy to clipboard")
			self.m_FileMenu.Append( self.m_ItemExport )
			self.Bind( wx.EVT_MENU, self.Export_OnMenu, id = self.m_ItemExport.GetId() )
		else:
			self.m_FileMenu = FileMenu

		self.m_menubar.Append( self.m_FileMenu, u"File" )

		self.m_menuDigits = wx.Menu()
		self.m_menuItem2 = wx.MenuItem( self.m_menuDigits, wx.ID_ANY, u"2 Digits", wx.EmptyString, wx.ITEM_RADIO )
		self.m_menuDigits.Append( self.m_menuItem2 )

		self.m_menuItem3 = wx.MenuItem( self.m_menuDigits, wx.ID_ANY, u"3 Digits", wx.EmptyString, wx.ITEM_RADIO )
		self.m_menuDigits.Append( self.m_menuItem3 )

		self.m_menuItem4 = wx.MenuItem( self.m_menuDigits, wx.ID_ANY, u"4 Digits", wx.EmptyString, wx.ITEM_RADIO )
		self.m_menuDigits.Append( self.m_menuItem4 )

		self.m_menuItem_AsIs = wx.MenuItem( self.m_menuDigits, wx.ID_ANY, u"As Is", wx.EmptyString, wx.ITEM_RADIO )
		self.m_menuDigits.Append( self.m_menuItem_AsIs )
		self.m_menuItem_AsIs.Check( True )

		self.m_menubar.Append( self.m_menuDigits, u"Digits" )
		self.SetMenuBar( self.m_menubar )


		self.Centre( wx.BOTH )

		
		self.Bind( wx.EVT_MENU, self.Digit2_OnMenu, id = self.m_menuItem2.GetId() )
		self.Bind( wx.EVT_MENU, self.Digit3_OnMenu, id = self.m_menuItem3.GetId() )
		self.Bind( wx.EVT_MENU, self.Digit4_OnMenu, id = self.m_menuItem4.GetId() )
		self.Bind( wx.EVT_MENU, self.AsIs_OnMenu, id = self.m_menuItem_AsIs.GetId() )

	def __del__( self ):
		pass

	
	def Export_OnMenu( self, event ):
		curPage = self.m_notebook.GetCurrentPage()
		curPage.Export()
		
		event.Skip()
	
	def Digit2_OnMenu( self, event ):
		self.m_Digits = 2
		event.Skip()

	def Digit3_OnMenu( self, event ):
		self.m_Digits = 3
		event.Skip()

	def Digit4_OnMenu( self, event ):
		self.m_Digits = 4
		event.Skip()

	def AsIs_OnMenu( self, event ):
		self.m_Digits = None
		event.Skip()

	def GetDigits(self):
            return self.m_Digits


import ctypes as _ct
from ..util import parent_path as _parent_path


_path = _parent_path(__file__) / "scisuit_core"
core = _ct.PyDLL(str(_path))





"""       ENGINEERING               """

core.c_eng_psychrometry.argtypes = [_ct.py_object]
core.c_eng_psychrometry.restype = _ct.py_object




"""        ROOTS PACKAGE           """

core.c_root_bisect.argtypes = [_ct.py_object, _ct.c_double, _ct.c_double, _ct.c_double, _ct.c_int, _ct.c_char_p, _ct.c_bool]
core.c_root_bisect.restype = _ct.py_object

core.c_root_brentq.argtypes = [_ct.py_object, _ct.c_double, _ct.c_double, _ct.c_double, _ct.c_int]
core.c_root_brentq.restype = _ct.py_object

core.c_root_muller.argtypes = [_ct.py_object, _ct.py_object, _ct.py_object, _ct.py_object, _ct.py_object, _ct.c_double, _ct.c_int]
core.c_root_muller.restype = _ct.py_object

core.c_root_newton.argtypes = [_ct.py_object, _ct.c_double, _ct.py_object, _ct.py_object, _ct.c_double, _ct.c_int]
core.c_root_newton.restype = _ct.py_object

core.c_root_ridder.argtypes = [_ct.py_object, _ct.c_double, _ct.c_double, _ct.c_double, _ct.c_int]
core.c_root_ridder.restype = _ct.py_object




"""                 FITTING PACKAGE"""

core.c_fit_expfit.argtypes = [_ct.py_object, _ct.py_object, _ct.py_object]
core.c_fit_expfit.restype = _ct.py_object
core.c_fit_lagrange.argtypes = [_ct.py_object, _ct.py_object,_ct.c_double]
core.c_fit_lagrange.restype = _ct.py_object
core.c_fit_logfit.argtypes = [_ct.py_object, _ct.py_object]
core.c_fit_logfit.restype = _ct.py_object
core.c_fit_logistfit.argtypes = [_ct.py_object, _ct.py_object, _ct.py_object]
core.c_fit_logistfit.restype = _ct.py_object
core.c_fit_powfit.argtypes = [_ct.py_object, _ct.py_object]
core.c_fit_powfit.restype = _ct.py_object
core.c_fit_spline.argtypes = [_ct.py_object, _ct.py_object]
core.c_fit_spline.restype = _ct.py_object



"""        STATS PACKAGE           """

core.c_stat_dbinom.argtypes = [_ct.py_object, _ct.c_int, _ct.c_double]
core.c_stat_dbinom.restype=_ct.py_object
core.c_stat_pbinom.argtypes = [_ct.py_object, _ct.c_int, _ct.c_double]
core.c_stat_pbinom.restype=_ct.py_object
core.c_stat_qbinom.argtypes = [_ct.py_object, _ct.c_int, _ct.c_double]
core.c_stat_qbinom.restype=_ct.py_object


core.c_stat_dchisq.argtypes = [_ct.py_object, _ct.c_int]
core.c_stat_dchisq.restype=_ct.py_object
core.c_stat_pchisq.argtypes = [_ct.py_object, _ct.c_int]
core.c_stat_pchisq.restype=_ct.py_object
core.c_stat_qchisq.argtypes = [_ct.py_object, _ct.c_int]
core.c_stat_qchisq.restype=_ct.py_object


core.c_stat_df.argtypes = [_ct.py_object, _ct.c_int, _ct.c_int]
core.c_stat_df.restype=_ct.py_object
core.c_stat_pf.argtypes = [_ct.py_object, _ct.c_int, _ct.c_int]
core.c_stat_pf.restype=_ct.py_object
core.c_stat_qf.argtypes = [_ct.py_object, _ct.c_int, _ct.c_int]
core.c_stat_qf.restype=_ct.py_object


core.c_stat_dnorm.argtypes = [_ct.py_object, _ct.c_double, _ct.c_double]
core.c_stat_dnorm.restype=_ct.py_object
core.c_stat_pnorm.argtypes = [_ct.py_object, _ct.c_double, _ct.c_double]
core.c_stat_pnorm.restype=_ct.py_object
core.c_stat_qnorm.argtypes = [_ct.py_object, _ct.c_double, _ct.c_double]
core.c_stat_qnorm.restype=_ct.py_object


core.c_stat_dpois.argtypes = [_ct.py_object, _ct.c_double]
core.c_stat_dpois.restype=_ct.py_object
core.c_stat_ppois.argtypes = [_ct.py_object, _ct.c_double]
core.c_stat_ppois.restype=_ct.py_object
core.c_stat_qpois.argtypes = [_ct.py_object, _ct.c_double]
core.c_stat_qpois.restype=_ct.py_object


core.c_stat_dt.argtypes = [_ct.py_object, _ct.c_int]
core.c_stat_dt.restype=_ct.py_object
core.c_stat_pt.argtypes = [_ct.py_object, _ct.c_int]
core.c_stat_pt.restype=_ct.py_object
core.c_stat_qt.argtypes = [_ct.py_object, _ct.c_int]
core.c_stat_qt.restype=_ct.py_object


core.c_stat_dunif.argtypes = [_ct.py_object, _ct.c_double, _ct.c_double]
core.c_stat_dunif.restype=_ct.py_object
core.c_stat_punif.argtypes = [_ct.py_object, _ct.c_double, _ct.c_double]
core.c_stat_punif.restype=_ct.py_object
core.c_stat_qunif.argtypes = [_ct.py_object, _ct.c_double, _ct.c_double]
core.c_stat_qunif.restype=_ct.py_object


core.c_stat_dsignrank.argtypes = [_ct.py_object, _ct.c_int]
core.c_stat_dsignrank.restype=_ct.py_object
core.c_stat_psignrank.argtypes = [_ct.py_object, _ct.c_int]
core.c_stat_psignrank.restype=_ct.py_object
core.c_stat_qsignrank.argtypes = [_ct.py_object, _ct.c_int]
core.c_stat_qsignrank.restype=_ct.py_object


core.c_stat_moveavg.argtypes = [_ct.py_object, _ct.py_object, _ct.c_int]
core.c_stat_moveavg.restype=_ct.py_object

core.c_stat_rolling.argtypes = [_ct.py_object, _ct.py_object, _ct.c_int]
core.c_stat_rolling.restype=_ct.py_object

core.c_stat_test_norm_ad.argtypes = [_ct.py_object]
core.c_stat_test_norm_ad.restype=_ct.py_object



__all__ = ['core']


import math
import numpy as np

from dataclasses import dataclass
from . import pnorm, qnorm



__all__ =['test_z', 'test_z_Result']




@dataclass
class test_z_Result:
	SE:float ; stdev:float 
	mean: float 
	zcritical:float 
	CI_upper: float ; CI_lower: float
	N:int


def test_z(x, sd, mu, alternative="two.sided", conflevel=0.95):
	"""
	
	## Return
	p-value and test_z_Result class. \n

	## Input
	x: sample,  ndarray/list \n
	sd: Standard deviation of population \n
	mu: Assumed mean of population \n
	alternative: "two.sided", "less", "greater" \n
	conflevel: Confidence level, (0,1)
	"""
	assert conflevel>=0.0 or conflevel <= 1.0, "conflevel must be in range (0, 1)"
	assert sd >= 0, "sd must be >0"
	assert isinstance(x, list) or type(x)==np.ndarray, "x must be list/ndarray"

	XX = x
	if isinstance(x, list):
		XX = np.asfarray(x)

	assert len(XX)>= 3, "container must have at least 3 elements"
	assert np.issubdtype(XX.dtype, np.number), "x must contain only numbers"

	dim = len(XX)
	xaver = np.mean(XX)
	SE = sd / math.sqrt(dim)
	stdeviation = np.std(XX, ddof =1); #sample's standard deviation

	alpha = 1 - conflevel
	zcritical = (xaver - mu) / SE

	pvalue = 0.0

	if alternative == "two.sided" or alternative == "notequal":
		if zcritical <= 0:
			#area on the left of zcrit + area on the right of positive
			pvalue = pnorm(zcritical, 0.0, 1.0) + (1.0 - pnorm(abs(zcritical), 0.0, 1.0))

		elif zcritical > 0:
			#area on the right of positive zcriticial + area on the left of negative zcriticial
			pvalue = (1.0 - pnorm(zcritical, 0.0, 1.0)) + pnorm(-zcritical, 0.0, 1.0)

		elif alternative == "greater":
			pvalue = (1.0 - pnorm(zcritical, 0.0, 1.0)) #area on the right

		elif alternative == "less":
			pvalue = pnorm(zcritical, 0.0, 1.0) #area on the left

		else:
			raise ValueError("Values for arg 'alternative': \"two.sided\" or \"notequal\", \"greater\", \"less\"")
	
	
	CI_upper = xaver - qnorm(alpha / 2.0, 0.0, 1.0) * SE
	CI_lower = xaver + qnorm(alpha / 2.0, 0.0, 1.0) * SE

	result = test_z_Result(SE = SE, 
			stdev = stdeviation, 
			N = dim, 
			mean = xaver, 
			zcritical = zcritical,
			CI_lower = CI_lower,
			CI_upper = CI_upper)

	return pvalue, result


import numpy as _np
import math as _math



def skew(y:_np.ndarray | list):
	"""
	Computes the skewness of a distribution

	## Example
	y = _np.random.standard_normal(1000)
	print(skew(y))
	
	"""
	n = len(y)

	assert n >= 3, "list/ndarray must have at least 3 elements"
	assert isinstance(y, list) or isinstance(y, _np.ndarray), "list/ndarray expected"
	
	Arr = None
	if(isinstance(y, list)):
		Arr = _np.asarray(y)
	else:
		Arr = _np.array(y)
	
	avg, std_p = _np.mean(Arr), _np.std(Arr)

	skew_p = _np.sum((Arr-avg)**3)
	skew_p /= (n * std_p**3.0);

	return _math.sqrt(n * (n - 1)) * skew_p/(n-2)



"""
y = _np.random.standard_normal(1000)
print(skew(y))
"""


import numpy as _np

def mode(y:_np.ndarray | list)->tuple:
	"""
	Computes mode

	## Example
	arr = np.array([1, 3, 5, 5, 7, 9, 3, 5, 7, 3, 6])
	print(mode(y = arr))
	"""
	n = len(y)

	assert n >= 3, "list/ndarray must have at least 3 elements"
	assert isinstance(y, list) or isinstance(y, _np.ndarray), "list/ndarray expected"
	
	Arr = None
	if(isinstance(y, list)):
		Arr = _np.asarray(y)
	else:
		Arr = _np.array(y)

	values, counts = _np.unique(Arr, return_counts=True)
	if(len(values) == len(Arr)):
		return ([], 0)

	index = _np.argwhere(counts == _np.max(counts))

	return values[index].flatten().tolist(), int(_np.max(counts))

"""
t1=[1, 2, 3, 4, 5]
t2=[1, 3, 5, 7, 9, 3, 5, 7, 3, 6]
arr = np.array([1, 3, 5, 5, 7, 9, 3, 5, 7, 3, 6])
print(mode(y = arr))
"""


import sys, os

#inserting to 0th position is very important so that search will FIRST match ../scisuit folder
sys.path.insert(0, os.getcwd()) 


from scisuit.eng import psychrometry

result = psychrometry(P=101, Tdb=30, Twb=20)

#all of the properties
print(result)

print("----------------------- \n")

#an individual property (enthalpy)
print("h=" + str(result.H))


import sys, os

#inserting to 0th position is very important so that search will FIRST match ../scisuit folder
sys.path.insert(0, os.getcwd()) 

from scisuit.eng.fluids_dbase import ThermoPhysical, SaturatedRefrigerant

f = ThermoPhysical("water")
res = f.search("T", 20)
print(res)

ref = SaturatedRefrigerant("water")
res = ref.search("T", 30)
print(res)


import numpy as _np

def sample(y, size:int, replace = False):
	"""
	y: ndarray / list.

	## Example
	ss = _np.linspace(start=1, stop=10, num=10) \n
	s = sample(ss, 4) \n
	s1 = sample(ss, 11, replace=True) \n
	s2 = sample(ss, 11, replace=False) \n
	print(s1)
	"""
	n = len(y)

	assert n >= 3, "list/ndarray must have at least 3 elements"
	assert isinstance(y, list) or isinstance(y, _np.ndarray), "list/ndarray expected"
	
	Arr = None
	if(isinstance(y, list)):
		Arr = _np.asarray(y)
	else:
		Arr = _np.array(y)

	if replace == False:
		if size>=n:
			raise ValueError("when replace=False, size < len(y) expected")
		_np.random.shuffle(Arr)
		
		return _np.array(Arr[0:size]).tolist()
	
	else:
		Index = _np.random.randint(low=0, high=n, size=size)
		t=[]
		for i in Index:
			t.append(Arr[i])
		return t

"""
ss = _np.linspace(start=1, stop=10, num=10)
s = sample(ss, 4)
s1 = sample(ss, 11, replace=True)
s2 = sample(ss, 11, replace=False)
print(s1)
"""


import sys, os
import numpy as np

#inserting to 0th position is very important so that search will FIRST match ../scisuit folder
sys.path.insert(0, os.getcwd()) 

from scisuit import trapz, cumtrapz

x = np.arange(1, 6)
y = x**2

print("****** cumtrapz ******")
print(cumtrapz(x, y))


print("****** trapz ******")

x=np.linspace(0, 3, num=100)
y = x**2

print(trapz(x,y))


from .._ctypeslib import core as _core

import numpy as np


__all__ = ['rolling']


class rolling:

	"""
	A class to compute rolling window mean, std, var min, max...

	## Example:
	>>r = rolling(x, y, period=2) \n
	>>r.mean() \n
	>>r.var() \n \n

	OR \n
	
	>>r=rolling(x, y, period=2).mean() \n

	If multiple calls required, prefer the first way for efficiency.

	"""

     
	def __init__(self, x:list, y:list, period:int = 2):
		assert isinstance(x, list) or isinstance(x, np.ndarray), "x must be of type list/ndarray"
		assert isinstance(y, list) or isinstance(y, np.ndarray), "y must be of type list/ndarray"
		assert isinstance(period, int), "period must be of type int"

		self._m_X, self._m_Windows = _core.c_stat_rolling(x=x, y=y, period=period)
		

	def __compute(self, func)->list:
		"""
		Calls the parameter func on self._m_Windows
		Returns a 1D list containing real numbers
		"""
		retList = []
		for Lst in self._m_Windows:
			retList.append(func(Lst))
		
		return retList
	

	def get(self):
		"""
		When mean, median, min, max member funcs do not meet the purpose.<br>
		To process rolling windows data (2D list) manually. <br>
		get()->(list, 2D list)
		"""
		return self._m_X, self._m_Windows
	

	def mean(self):
		return self.__compute(self.__mean)

	def median(self):
		return self.__compute(self.__median)
	
	def min(self):
		return self.__compute(min)
	
	def max(self):
		return self.__compute(max)

	

	def __mean(self, lst:list):
		return sum(lst)/len(lst)

	def __median(self, arg):
		arr:np.ndarray = arg

		if isinstance(arg, list):
			arr = np.asfarray(arg)
		
		return float(np.median(arr))


	


import numpy as _np
import math as _math

def trapz(x:list|_np.ndarray, y:list|_np.ndarray)->float:
    """
    Computes area using trapezoidal method and uses Numpy's trapz method.

    ## Input: 
    x, y: list or ndarray
    """
    return _np.trapz(x=x, y=y)



def cumtrapz(x:list|_np.ndarray, y:list|_np.ndarray)->list:
	"""
	Computes the left-tailed cumulative area

	##Input:
	x, y: list or ndarray
	"""
	val = 0.0
	a, b, f_a, f_b = 0.0, 0.0, 0.0, 0.0
	retList =[0.0]

	for i in range(len(x)-1):
		a = x[i]
		b = x[i + 1]

		if _math.isclose(b, a, abs_tol = 1E-5):
			raise ValueError("X data contains successive entries whose difference is smaller than 1E-5")

		if (b < a):
			raise ValueError("X data is not sorted in ascending order.")

		f_a = y[i]
		f_b = y[i + 1]

		val += (b - a) * (f_a + f_b) / 2.0

		retList.append(val)
	
	return retList


import math
import numpy as np

from . import pt, qt

from dataclasses import dataclass


__all__ = ['test_t', 'test_t1_result', 'test_t2_result', 'test_tpaired_result']



@dataclass
class test_t1_result:
	CI_lower:float; CI_upper:float
	SE:float
	N:int
	stdev:float
	mean: float
	tcritical:float


@dataclass
class test_t2_result:
	CI_lower:float; CI_upper:float
	tcritical:float
	n1:int; n2:int
	df:int
	xaver:float; yaver:float
	s1:float; s2:float; sp:float


@dataclass
class test_tpaired_result:
	CI_lower:float; CI_upper:float
	tcritical:float
	xaver:float; yaver:float
	s1:float; s2:float; SE:float
	N:int
	mean:float #mean of difference
	stdev:float #stdev of difference


def _test_t1(x, mu, alternative="two.sided", conflevel=0.95):
	assert conflevel>=0.0 or conflevel <= 1.0, "conflevel must be in range (0, 1)"
	assert isinstance(x, list) or type(x)==np.ndarray, "x must be list/ndarray"

	XX = x
	if isinstance(x, list):
		XX = np.asfarray(x)

	assert len(XX)>= 3, "container must have at least 3 elements"
	assert np.issubdtype(XX.dtype, np.number), "x must contain only numbers"

	N = len(XX)
	df = N -1
	
	stdev = np.std(XX, ddof =1) #sample's standard deviation
	SE = stdev / math.sqrt(N) #Standard Error of Mean
	
	xaver = np.mean(XX)
	tcritical = (xaver - mu) / SE

	pvalue = 0.0
	if alternative == "two.sided" or alternative == "notequal":
		if (tcritical <= 0.0):
			#area on the left of tcrit + area on the right of positive
			pvalue = pt(tcritical, df) + (1.0 - pt(abs(tcritical), df))
		else:
			#area on the right of positive tcritical + area on the left of negative tcritical
			pvalue = (1.0 - pt(tcritical, df)) + pt(-tcritical, df)
	
	#area on the right
	elif alternative == "greater":
		pvalue = (1.0 - pt(tcritical, df))
	
	#area on the left
	elif alternative == "less":
		pvalue = pt(tcritical, df); 

	else:
		raise ValueError("Values for 'alternative': \"two.sided\" or \"notequal\", \"greater\", \"less\"")
	
	alpha = 1.0 - conflevel
	CI_upper = xaver - qt(alpha / 2.0, df) * SE
	CI_lower = xaver + qt(alpha / 2.0, df) * SE

	Result = test_t1_result(
		CI_lower=CI_lower,
		CI_upper=CI_upper,
		SE = SE,
		N = N,
		stdev = stdev,
		mean = xaver,
		tcritical = tcritical)
	
	return pvalue, Result



def _test_t2(x, y, mu, varequal = True, alternative="two.sided", conflevel=0.95):
	assert conflevel>=0.0 or conflevel <= 1.0, "conflevel must be in range (0, 1)"
	assert isinstance(x, list) or type(x)==np.ndarray, "x must be list/ndarray"

	XX = x
	if isinstance(x, list):
		XX = np.asfarray(x)
	
	YY = y
	if isinstance(y, list):
		YY = np.asfarray(y)

	assert len(XX)>= 3, "x must have at least 3 elements"
	assert len(YY)>= 3, "y must have at least 3 elements"
	assert np.issubdtype(XX.dtype, np.number), "x must contain only numbers"
	assert np.issubdtype(YY.dtype, np.number), "y must contain only numbers"

	n1, n2 = len(XX), len(YY)
	xaver, yaver = np.mean(XX), np.mean(YY)
	s1, s2 = np.std(XX, ddof = 1), np.std(YY, ddof = 1)
	var1, var2 = s1**2, s2**2
	alpha = 1 - conflevel

	tcritical, SE = None, None
	sp = -1 #pooled

	if varequal == False:
		df_num = (var1 / n1 + var2 / n2)**2
		df_denom = 1 / (n1 - 1) * (var1 / n1)**2 + 1 / (n2 - 1) * (var2 / n2)**2
		df = math.floor(df_num / df_denom)

		tcritical = ((xaver - yaver) - mu) / math.sqrt(var1 / n1 + var2 / n2)
		SE = math.sqrt(var1 / n1 + var2 / n2)
	
	else:
		df = n1 + n2 - 2

		sp_num = (n1 - 1) * var1 + (n2 - 1) * var2
		sp = math.sqrt(sp_num / df)

		tcritical = ((xaver - yaver) - mu) / (sp * math.sqrt(1 / n1 + 1 / n2))
		SE = sp * math.sqrt(1 / n1 + 1 / n2)
	
	pvalue = 0
	if alternative == "two.sided" or alternative == "notequal":
		if tcritical <= 0:
			#area on the left of tcrit + area on the right of positive
			pvalue = pt(tcritical, df) + (1 - pt(abs(tcritical), df))
		else:
			#area on the right of positive tcritical + area on the left of negative tcritical
			pvalue = (1 - pt(tcritical, df)) + pt(-tcritical, df)
	
	# area on the right
	elif alternative == "greater":	
		pvalue = 1.0 - pt(tcritical, df)
	
	# area on the left
	elif alternative == "less":	
		pvalue = pt(tcritical, df)
	
	else:
		raise ValueError("Values for 'alternative': \"two.sided\" or \"notequal\", \"greater\", \"less\"")
	
	quantile = qt(alpha / 2, df)

	CI1 = (xaver - yaver) - quantile * SE
	CI2 = (xaver - yaver) + quantile * SE

	CI_lower = min(CI1, CI2)
	CI_upper = max(CI1, CI2)

	Result = test_t2_result(
		CI_lower=CI_lower, CI_upper=CI_upper,
		tcritical=tcritical,
		n1 = n1, n2= n2,
		df=df,
		s1 = s1, s2=s2, sp=sp,
		xaver = xaver, yaver = yaver)
	
	return pvalue, Result


def _test_t_paired(x, y, mu, alternative="two.sided", conflevel=0.95):
	assert conflevel>=0.0 or conflevel <= 1.0, "conflevel must be in range (0, 1)"
	assert isinstance(x, list) or type(x)==np.ndarray, "x must be list/ndarray"

	XX = x
	if isinstance(x, list):
		XX = np.asfarray(x)
	
	YY = y
	if isinstance(y, list):
		YY = np.asfarray(y)

	assert len(XX)>= 3, "container must have at least 3 elements"
	assert np.issubdtype(XX.dtype, np.number), "x must contain only numbers"
	assert np.issubdtype(YY.dtype, np.number), "y must contain only numbers"
	assert len(XX) == len(YY), "x and y must have same size"

	xaver, yaver = np.mean(XX), np.mean(YY)
	s1, s2 = np.std(XX, ddof = 1), np.std(YY, ddof = 1)
	
	Diff = XX- YY
	pvalue, Res1 = _test_t1(x =Diff, mu = mu, alternative = alternative, conflevel = conflevel)

	Result = test_tpaired_result(
		CI_lower=Res1.CI_lower,
		CI_upper = Res1.CI_upper,
		tcritical= Res1.tcritical,
		xaver=xaver,
		yaver=yaver,
		s1=s1,
		s2= s2,
		SE = Res1.SE,
		mean=Res1.mean,
		N = Res1.N,
		stdev=Res1.stdev)
	
	return pvalue, Result



def test_t (x, y = None, varequal=True, alternative="two.sided", mu=0, conflevel=0.95, paired=False ):
	"""
	Performs paired, 1-sample and 2-sample t-test
	"""
	if y == None:
		return _test_t1(x=x, mu=mu, alternative=alternative, conflevel=conflevel )	
	else:
		if paired==False:
			return _test_t2(x=x, y=y, mu=mu, varequal=varequal, alternative=alternative, conflevel=conflevel)
		else:
			return _test_t_paired(x=x, y=y, mu=mu, alternative=alternative, conflevel=conflevel)


import sys, os

#inserting to 0th position is very important so that search will FIRST match ../scisuit folder
sys.path.insert(0, os.getcwd()) 


from scisuit.eng import Food

f1= Food (water =80, CHO=20)
f2= Food (water =60, CHO=10, protein = 30)

f1.T = 10 #at 10°C
f2.T = 25 #at 25°C

"""create a mixture of equal weights
mass and energy balances performed automatically"""
f3 = f1 + f2

print(f3)

"""
milk = Food(
    water=88.13, 
    protein=3.15, 
    CHO=4.80, 
    lipid=3.25, 
    ash=0.67)

water = Food(water=100)

#removal of 87% water from milk
powder = milk - 0.87*water 

print(milk)

print("------------ \n")

print(powder)

"""



import time
import wx


import scisuit.eng as eng
from scisuit.util import parent_path
from scisuit.wxpy import makeicon, NumTextCtrl





class frmPsychrometry ( wx.Frame ):

	def __init__( self, parent = None, FileMenu = None ):
		wx.Frame.__init__ ( self, 
                parent,  	
                title = u"Psychrometry", 
                style = wx.CAPTION|wx.CLOSE_BOX|wx.MINIMIZE_BOX|wx.RESIZE_BORDER|wx.TAB_TRAVERSAL )

		self.m_Digits = 3 #show results with 3 decimal points

		IconPath = parent_path(__file__) / "icons" / "psychart.png"
		self.SetIcon(makeicon(IconPath))

		self.SetSizeHints( wx.DefaultSize, wx.DefaultSize )
		self.SetBackgroundColour( wx.Colour( 0, 242, 242 ) )

		self.m_chkP = wx.CheckBox( self, wx.ID_ANY, "P")
		self.m_chkP.SetToolTip( u"Pressure" )
		self.m_txtP = NumTextCtrl( self)
		self.m_txtP.SetToolTip( u"Pressure" )
		self.m_lblP = wx.StaticText( self, wx.ID_ANY, "kPa")
		self.m_lblP.Wrap( -1 )

		self.m_stxtPw = wx.StaticText( self, wx.ID_ANY, "Pw")
		self.m_stxtPw.Wrap( -1 )
		self.m_txtPw = wx.TextCtrl( self, style= wx.TE_READONLY )
		self.m_lblPw = wx.StaticText( self, wx.ID_ANY, "kPa")
		self.m_lblPw.Wrap( -1 )

		self.m_stxtPws = wx.StaticText( self, wx.ID_ANY, u"Pws")
		self.m_stxtPws.Wrap( -1 )
		self.m_txtPws = wx.TextCtrl( self, style= wx.TE_READONLY )
		self.m_lblPws = wx.StaticText( self, wx.ID_ANY, "kPa")
		self.m_lblPws.Wrap( -1 )

		self.m_chkTdb = wx.CheckBox( self, wx.ID_ANY, u"Tdb")
		self.m_chkTdb.SetToolTip( u"dry-bulb temperature" )
		self.m_txtTdb = NumTextCtrl( self)
		self.m_lblTdb = wx.StaticText( self, wx.ID_ANY, "°C")
		self.m_lblTdb.Wrap( -1 )

		self.m_chkTwb = wx.CheckBox( self, wx.ID_ANY, u"Twb")
		self.m_chkTwb.SetToolTip( u"wet-bulb temperature" )
		self.m_txtTwb = NumTextCtrl( self)
		self.m_lblTwb = wx.StaticText( self, wx.ID_ANY, "°C")
		self.m_lblTwb.Wrap( -1 )

		self.m_chkTdp = wx.CheckBox( self, wx.ID_ANY, u"Tdp")
		self.m_chkTdp.SetToolTip( u"dew-point temperature" )
		self.m_txtTdp = NumTextCtrl( self)
		self.m_lblTdp = wx.StaticText( self, wx.ID_ANY, "°C")
		self.m_lblTdp.Wrap( -1 )

		self.m_chkW = wx.CheckBox( self, wx.ID_ANY, "W")
		self.m_txtW = NumTextCtrl( self)	
		self.m_lblW = wx.StaticText( self, wx.ID_ANY, "kg/kgda")
		self.m_lblW.Wrap( -1 )

		self.m_stxtWs = wx.StaticText( self, wx.ID_ANY, "Ws")
		self.m_stxtWs.Wrap( -1 )	
		self.m_txtWs = wx.TextCtrl( self,style = wx.TE_READONLY )
		self.m_lblWs = wx.StaticText( self, wx.ID_ANY, "kg/kgda")
		self.m_lblWs.Wrap( -1 )	

		self.m_chkH = wx.CheckBox( self, wx.ID_ANY, "H")
		self.m_txtH =  NumTextCtrl( self)
		self.m_lblH = wx.StaticText( self, wx.ID_ANY, "kJ/kgda")
		self.m_lblH.Wrap( -1 )

		self.m_chkRH = wx.CheckBox( self, wx.ID_ANY, "RH")
		self.m_txtRH = NumTextCtrl( self)
		self.m_lblRH = wx.StaticText( self, wx.ID_ANY, "%")
		self.m_lblRH.Wrap( -1 )

		self.m_chkV = wx.CheckBox( self, wx.ID_ANY, "v")
		self.m_txtV = NumTextCtrl( self)
		self.m_lblV = wx.StaticText( self, wx.ID_ANY, "m3/kg")
		self.m_lblV.Wrap( -1 )

		fgSzr_L = wx.FlexGridSizer( 0, 3, 0, 0 )
		fgSzr_L.SetFlexibleDirection( wx.BOTH )
		fgSzr_L.SetNonFlexibleGrowMode( wx.FLEX_GROWMODE_SPECIFIED )
		fgSzr_L.Add( self.m_chkP, 0, wx.ALL, 5 )
		fgSzr_L.Add( self.m_txtP, 0, wx.ALL, 5 )
		fgSzr_L.Add( self.m_lblP, 0, wx.ALL, 5 )
		fgSzr_L.Add( self.m_stxtPw, 0, wx.ALL, 5 )
		fgSzr_L.Add( self.m_txtPw, 0, wx.ALL, 5 )
		fgSzr_L.Add( self.m_lblPw, 0, wx.ALL, 5 )
		fgSzr_L.Add( self.m_stxtPws, 0, wx.ALL, 5 )
		fgSzr_L.Add( self.m_txtPws, 0, wx.ALL, 5 )
		fgSzr_L.Add( self.m_lblPws, 0, wx.ALL, 5 )
		fgSzr_L.Add( self.m_chkTdb, 0, wx.ALL, 5 )
		fgSzr_L.Add( self.m_txtTdb, 0, wx.ALL, 5 )
		fgSzr_L.Add( self.m_lblTdb, 0, wx.ALL, 5 )
		fgSzr_L.Add( self.m_chkTwb, 0, wx.ALL, 5 )
		fgSzr_L.Add( self.m_txtTwb, 0, wx.ALL, 5 )
		fgSzr_L.Add( self.m_lblTwb, 0, wx.ALL, 5 )
		fgSzr_L.Add( self.m_chkTdp, 0, wx.ALL, 5 )
		fgSzr_L.Add( self.m_txtTdp, 0, wx.ALL, 5 )
		fgSzr_L.Add( self.m_lblTdp, 0, wx.ALL, 5 )

		fgSzr_R = wx.FlexGridSizer( 0, 3, 0, 0 )
		fgSzr_R.SetFlexibleDirection( wx.BOTH )
		fgSzr_R.SetNonFlexibleGrowMode( wx.FLEX_GROWMODE_SPECIFIED )
		fgSzr_R.Add( self.m_chkW, 0, wx.ALL, 5 )
		fgSzr_R.Add( self.m_txtW, 0, wx.ALL, 5 )
		fgSzr_R.Add( self.m_lblW, 0, wx.ALL, 5 )
		fgSzr_R.Add( self.m_stxtWs, 0, wx.ALL, 5 )
		fgSzr_R.Add( self.m_txtWs, 0, wx.ALL, 5 )
		fgSzr_R.Add( self.m_lblWs, 0, wx.ALL, 5 )
		fgSzr_R.Add( self.m_chkH, 0, wx.ALL, 5 )
		fgSzr_R.Add( self.m_txtH, 0, wx.ALL, 5 )
		fgSzr_R.Add( self.m_lblH, 0, wx.ALL, 5 )
		fgSzr_R.Add( self.m_chkRH, 0, wx.ALL, 5 )
		fgSzr_R.Add( self.m_txtRH, 0, wx.ALL, 5 )
		fgSzr_R.Add( self.m_lblRH, 0, wx.ALL, 5 )
		fgSzr_R.Add( self.m_chkV, 0, wx.ALL, 5 )
		fgSzr_R.Add( self.m_txtV, 0, wx.ALL, 5 )
		fgSzr_R.Add( self.m_lblV, 0, wx.ALL, 5 )

		szrLR = wx.BoxSizer( wx.HORIZONTAL )
		szrLR.Add( fgSzr_L, 1, wx.EXPAND, 5 )
		szrLR.Add( fgSzr_R, 1, wx.EXPAND, 5 )

		self.m_btnCalc = wx.Button( self, wx.ID_ANY, "Compute")
		self.m_btnCalc.Enabled=False
		
		mainSizer = wx.BoxSizer( wx.VERTICAL )
		mainSizer.Add( szrLR, 1, wx.EXPAND, 5 )
		mainSizer.Add( ( 0, 20), 0, wx.EXPAND, 5 )
		mainSizer.Add( self.m_btnCalc, 0, wx.ALL|wx.ALIGN_CENTER_HORIZONTAL, 5 )
		
		if(FileMenu == None):
			self.m_menuFile = wx.Menu()
			self.m_menuCopyClipbrd = wx.MenuItem( self.m_menuFile, -1, "Copy to clipboard")
			self.m_menuFile.Append( self.m_menuCopyClipbrd )
			self.Bind( wx.EVT_MENU, self.__OnCopyClipbrd, id = self.m_menuCopyClipbrd.GetId() )
		else:
			assert isinstance(FileMenu, wx.Menu), "FileMenu must be of type wx.Menu"
			self.m_menuFile = FileMenu
		
		self.m_menuDigits = wx.Menu()	
		self.m_menuItem2Digits = wx.MenuItem( self.m_menuDigits, -1, "2 Digits","", wx.ITEM_RADIO)
		self.m_menuDigits.Append( self.m_menuItem2Digits )

		self.m_menuItem3Digits = wx.MenuItem( self.m_menuDigits, -1, "3 Digits", "", wx.ITEM_RADIO )
		self.m_menuDigits.Append( self.m_menuItem3Digits )
		self.m_menuItem3Digits.Check(True)

		self.m_menuItem4Digits = wx.MenuItem( self.m_menuDigits, -1, "4 Digits", "", wx.ITEM_RADIO )
		self.m_menuDigits.Append( self.m_menuItem4Digits )

		self.m_menubar = wx.MenuBar( 0 )
		self.m_menubar.Append( self.m_menuFile, "File" )
		self.m_menubar.Append( self.m_menuDigits, "Digits" )
		self.SetMenuBar( self.m_menubar )


		self.SetSizerAndFit( mainSizer )
		self.Layout()
		self.Centre( wx.BOTH )


		self.Bind( wx.EVT_CHECKBOX, self.__OnCheckBox )
		self.m_btnCalc.Bind( wx.EVT_BUTTON, self.__OnBtn )
		
		
		self.Bind( wx.EVT_MENU, self.__OnMenuDigits, id = self.m_menuItem2Digits.GetId() )
		self.Bind( wx.EVT_MENU, self.__OnMenuDigits, id = self.m_menuItem3Digits.GetId() )
		self.Bind( wx.EVT_MENU, self.__OnMenuDigits, id = self.m_menuItem4Digits.GetId() )


		self.m_CheckBoxes=[self.m_chkP, self.m_chkTdb, self.m_chkTwb, self.m_chkTdp, 
				self.m_chkW, self.m_chkH, self.m_chkRH, self.m_chkV]

		
		self.m_Controls=[
			[self.m_chkP, self.m_txtP, "kPa", "P"],
			[self.m_chkTdb, self.m_txtTdb, "°C", "Tdb"],
			[self.m_chkTwb, self.m_txtTwb, "°C", "Twb"],
			[self.m_chkTdp, self.m_txtTdp, "°C", "Tdp"],
			[self.m_chkW, self.m_txtW, "kg/kg da", "W"],
			[self.m_chkH, self.m_txtH, "kJ/kgda", "H"],
			[self.m_chkRH, self.m_txtRH, "%", "RH"],
			[self.m_chkV, self.m_txtV, "m3/kg", "V"],
			[None, self.m_txtPw, "kPa", "Pw"],
			[None, self.m_txtPws, "kPa", "Pws"],
			[None, self.m_txtWs, "kg/kgda", "Ws"]
		]

	
	
	def EnableAllCheckBoxes(self):
		for chkBox in self.m_CheckBoxes:
			chkBox.Enable(True)
    

	def DisableUncheckedBoxes(self):
		for chkBox in self.m_CheckBoxes:
			chkBox.Enabled = chkBox.GetValue()



	def __del__( self ):
		pass

    
	def __OnCheckBox( self, event ):
		NChkedBoxes = 0
	
		for chkBox in self.m_CheckBoxes:
			NChkedBoxes += int(chkBox.GetValue())

		if(NChkedBoxes >= 3):
			self.DisableUncheckedBoxes()
		else:
			self.EnableAllCheckBoxes()
		
		self.m_btnCalc.Enabled = (NChkedBoxes >= 3)
		
	

	def __OnCopyClipbrd( self, event ):
		strInput, strOutput = "", ""
		
		for Entry in self.m_Controls:
			if(Entry[0] and Entry[0].GetValue()):
				strInput += str(Entry[3]) + " = " + str(Entry[1].GetValue()) + " " + str(Entry[2]) + "\n"
			else:
				strOutput += str(Entry[3]) + " = " + str(Entry[1].GetValue()) + " " + str(Entry[2]) + "\n"			

		data = strInput + "\n" + strOutput
		if wx.TheClipboard.Open():
			wx.TheClipboard.SetData(wx.TextDataObject(data))
			wx.TheClipboard.Close()
			wx.TheClipboard.Flush()
		
			
	
	
	def __OnMenuDigits( self, event ): 
		id = event.GetId()
		if(id == self.m_menuItem2Digits.GetId()):
			self.m_Digits=2
		elif(id == self.m_menuItem3Digits.GetId()):
			self.m_Digits=3
		else:
			self.m_Digits=4



	def __OnBtn( self, event ):
		PsyParams = dict()
		try:
			"""
			Collect the check entries and values in a dictionary
			"""
			for Entry in self.m_Controls:
				if(Entry[0] and Entry[0].GetValue()):
					assert Entry[1].GetValue() !="", "A numeric value must be entered for " + Entry[3]  

					#note that we use NumTextCtrl so this will always succeed
					PsyParams[Entry[3]] = float(Entry[1].GetValue()) 
		
		
			result = eng.psychrometry(**PsyParams)

			for Entry in self.m_Controls:
				value = getattr(result, Entry[3])
				value = round(value, self.m_Digits)
				Entry[1].SetValue(str(value))

		except Exception as e:
			wx.MessageBox(str(e))